///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  OverlapAndAddFiltering.cpp                                                        //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//
//<LIBTAG:9749d6b8-f609-4b79-bdff-84550a04d39d>

#include  "modules.h"
#include  "Exception.h"
#include  "{VAL params/FilterCoefDef}.h"
#include  "{VAL label}.h"


extern  mutex  FFTW_Mut;


{VAL label}::{VAL label}(void) :
StartUpState(e_InitState),
InData({VAL params/input_rfifo}),
FRT_In(InData),
FirFilter({VAL params/FilterCoefFifo}),
FRT_Coefs(FirFilter),
PostFilt_Dimensions  { {VAL params/output_fifo_length}, {VAL params/num_filters} },
PostFilt_WrapFlags  { true, false },
PostFilt_Extent  { PostFilt_Dimensions, PostFilt_WrapFlags },
{VAL params/output_fifo_label}(PostFilt_Extent),
FWT_Out({VAL params/output_fifo_label}),
FilterStartUpLag(0),
CurrWriteAmount(0),
READ(0),
FiltLen(0),
FftLen(0)
{
  this->State = State_Initialized;

  if( { VAL debug } )
  { // Debug display Display_Constructor
    cout << this->Display_Constructor();
  }
}


bool  {VAL label}::AnyActiveThreads(void) const
{
  bool  AnyActiveThreads = false;

  AnyActiveThreads = AnyActiveThreads || this->ControlThread.joinable();

  for( auto K=0U ; K < this->NumProcThreads ; ++K )
  {
    AnyActiveThreads = AnyActiveThreads || this->Thread[K].joinable();
  }

  return  AnyActiveThreads;
}


RTIF::ExitStateType  {VAL label}::Init_Allocate(void)
{
  this->State = State_Allocated;

  // Calculate 'NumProcThreads'
  if( this->NumProcThreads_FromParams <= this->NumChannels )
  {
    this->NumProcThreads =  this->NumProcThreads_FromParams;
  }
  else
  { // No need to spawn more threads than channels
    this->NumProcThreads = this->NumChannels;
  }

  // We need this value for allocating maximum array sizes during processing
  this->MaxChannelsPerProcThread = 0;

  // Determine which data channels are processed for each thread
  for( uint32_t K=0U ; K < this->NumProcThreads ; ++K )
  {
    this->BaseChannelNumberForThread( K,
                                      this->NumProcThreads,
                                      this->NumChannels,
                                      this->ChannelStart[K],
                                      this->NumChannelsPerThread[K] );

    if( this->NumChannelsPerThread[K] > this->MaxChannelsPerProcThread )
    {
      this->MaxChannelsPerProcThread = this->NumChannelsPerThread[K];
    }
  }

  // Set up control arrays
  this->FftLen.resize( this->NumChannels );

  for( auto K=0U ; K < this->NumChannels ; ++K )
  {
    this->FftLen[K] = {VAL params/GetFftLen}();
  }

  this->FiltLen.resize( this->NumChannels );

  for( auto K=0U ; K < this->NumChannels ; ++K )
  {
    this->FiltLen[K] = {VAL params/NumFilterCoefsCall}();
  }

  this->FilterStartUpLag.resize( this->NumChannels );

  for( auto K=0U ; K < this->NumChannels ; ++K )
  {
    this->FilterStartUpLag[K] = this->FiltLen[K] - 1;
  }

  this->READ.resize( this->NumChannels );

  for( auto K=0U ; K < this->NumChannels ; ++K )
  {
    this->READ[K] = this->FftLen[K] - this->FilterStartUpLag[K];
  }

  // If there are any currently active threads kill them and restart
  if( this->AnyActiveThreads() )
  { // If there are active threads: kill them
    this->KillThreads();
    this->KillFlag = false;
    this->State = State_Allocated;
  }

  // Reset some control variables
  this->Alive         = 0;
  this->Complete      = 0;
  this->CycleCounter  = ~0;
  this->StartUpState  = this->e_InitState;

  if( { VAL debug } )
  { // Display Display_InitAlloc
    cout << Display_InitAlloc();
  }

  return  RTIF::Exit_Normally;
}


RTIF::ExitStateType  {VAL label}::Init_Reset(void)
{
  this->State = State_Reset;

  // If there are any currently active threads kill them and restart
  if( this->AnyActiveThreads() )
  { // There are active threads: kill them
    this->KillThreads();
    this->KillFlag = false;
    this->State = State_Reset;
  }

  // Reset some control variables
  this->Alive         = 0;
  this->Complete      = 0;
  this->CycleCounter  = ~0;
  this->StartUpState  = this->e_InitState;

  // Set up control arrays
  this->FftLen.resize( this->NumChannels );

  for( auto K=0U ; K < this->NumChannels ; ++K )
  {
    this->FftLen[K] = {VAL params/GetFftLen}();
  }

  this->FiltLen.resize( this->NumChannels );

  for( auto K=0U ; K < this->NumChannels ; ++K )
  {
    this->FiltLen[K] = {VAL params/NumFilterCoefsCall}();
  }

  this->FilterStartUpLag.resize( this->NumChannels );

  for( auto K=0U ; K < this->NumChannels ; ++K )
  {
    this->FilterStartUpLag[K] = this->FiltLen[K] - 1;
  }

  this->READ.resize( this->NumChannels );

  for( auto K=0U ; K < this->NumChannels ; ++K )
  {
    this->READ[K] = this->FftLen[K] - this->FilterStartUpLag[K];
  }

  // Spawn the controller thread
  this->ControlThread = thread( this->ProcessBatches, ref(*this) );

  // Spawn the processing threads
  for( uint32_t K=0U ; K < this->NumProcThreads ; ++K )
  {
    this->ThreadCycleCounter[K] = ~0;
    this->Thread[K]             = thread( this->ProcessingThread, ref(*this), K );
  }

  { // Wait for all the threads to wake up and then wait on CycleCounter
    unique_lock<mutex>    LK(this->GetCvMutex());
    this->GetCv().wait( LK, [&]{ return (this->Alive == this->NumProcThreads) || this->KillFlag; } );
  }

  if( { VAL debug } )
  { // Display Display_InitReset
    cout << Display_InitReset();
  }

  return  RTIF::Exit_Normally;
}


// Assuming all filters have the same number of coefficients
RTIF::ExitStateType  {VAL label}::ProcessBatches( {VAL label}& _Obj )
{
  // Wait till we get a set of coefficients
  {
    unique_lock<mutex>  LK({VAL params/FilterCoefDec}.GetCvMutex());

    {VAL params/FilterCoefDec}.GetCv().wait( LK, [&]{
        return {VAL params/FilterCoefDec}.FilterGenerationComplete.load() || _Obj.KillFlag; } );
    _Obj.StartUpState = _Obj.e_GotFilterCoefs;
    LK.unlock();
    _Obj.GetCv().notify_all();
  }

  // Look for a Kill signal
  if( _Obj.KillFlag )  return  RTIF::Exit_Due_To_Kill;

  // READ FILTER COEFFICIENTS
  _Obj.FirFilter.FIFO_Read
  (
    _Obj.FirFilter.READ_FIXED_AMOUNT,
    _Obj.FftLen[0],
    _Obj.FRT_Coefs
  );

  // Notify the processing threads
  {
    unique_lock<mutex>   LK(_Obj.GetCvMutex());
    _Obj.StartUpState = _Obj.e_ReadFilterCoefs;
    LK.unlock();
    _Obj.GetCv().notify_all();
  }

  // Done writing into FIFO --- alert readers new space is available
  _Obj.FirFilter.FIFO_Read_End( _Obj.FRT_Coefs );

  /////////////////////////
  // Main Processing Loop
  /////////////////////////

  do
  {
    // Determine the amount to write
    _Obj.CurrWriteAmount = _Obj.READ[0];

    if( _Obj.CurrWriteAmount < _Obj.FilterStartUpLag[0] )
    {
      _Obj.CurrWriteAmount = 0;
    }
    else
    {
      _Obj.CurrWriteAmount -= _Obj.FilterStartUpLag[0];
    }

    // Wait for output space
    _Obj.{VAL params/output_fifo_label}.FIFO_Write_Start( _Obj.CurrWriteAmount, _Obj.FWT_Out );

    _Obj.InData.FIFO_Read
    (
      _Obj.InData.READ_FIXED_AMOUNT_EXCEPT_ON_LAST,
      _Obj.READ[0],
      _Obj.FRT_In
    );

    // Look for a Kill signal
    if( _Obj.KillFlag )  break;

    { // Increment the cycle count to start processing
      unique_lock<mutex>   LK(_Obj.GetCvMutex());
      ++_Obj.CycleCounter;
      _Obj.Complete = 0;
      LK.unlock();
      _Obj.GetCv().notify_all();
    }

    // Look for a Kill signal
    if( _Obj.KillFlag )  break;

    /////////////////////////////////
    // Processing occurs here
    /////////////////////////////////

    { // Wait for children to finish
      unique_lock<mutex>   LK(_Obj.GetCvMutex());
      _Obj.GetCv().wait( LK, [&]{ return (_Obj.Complete == _Obj.NumProcThreads) || _Obj.KillFlag; } );
    }

    // Look for a Kill signal
    if( _Obj.KillFlag )  break;

    // Done writing into FIFO --- alert readers new space is available
    _Obj.{VAL params/output_fifo_label}.FIFO_Write_End(_Obj.FWT_Out);

    // Done writing into FIFO --- alert readers new space is available
    _Obj.InData.FIFO_Read_End(_Obj.FRT_In);

    // Update the Lag counters
    for( auto K=0U ; K < _Obj.NumChannels ; ++K )
    {
      if( _Obj.CurrWriteAmount < _Obj.FilterStartUpLag[K] )
      {
        _Obj.FilterStartUpLag[K] -= (_Obj.READ[0] - _Obj.CurrWriteAmount);
      }
      else
      {
        _Obj.FilterStartUpLag[K] = 0;
      }
    }
  }  while( !_Obj.KillFlag );

  // Propagate the EndOfData flag downward
  if( !_Obj.KillFlag )
  {
    _Obj.{VAL params/output_fifo_label}.SetEndOfDataFlag();
  }

  return  RTIF::Exit_Due_To_Kill;
}


RTIF::ExitStateType  {VAL label}::ProcessingThread( {VAL label}& _Obj, const uint32_t _ThreadNum )
{
  uint32_t  ChanOffset;
  const vector<uint32_t>  LocVect_In_Incr = { 1, 0 };
  const vector<uint32_t>  LocVect_CoefsIn_Incr = { 1, 0 };
  const vector<uint32_t>  LocVect_Out_Incr = { 1, 0 };
  vector<uint32_t>        LocVect_In = vector<uint32_t>(_Obj.InData.GetVectorLen());
  vector<uint32_t>        LocVect_CoefsIn = vector<uint32_t>(_Obj.FirFilter.GetVectorLen());
  vector<uint32_t>        LocVect_Out = vector<uint32_t>(_Obj.{VAL params/output_fifo_label}.GetVectorLen());

  // Grab local processing parameters
  _Obj.GetCvMutex().lock();
  const auto    LocChanBase   = _Obj.ChannelStart[_ThreadNum];
  const auto    LocNumChans   = _Obj.NumChannelsPerThread[_ThreadNum];
  _Obj.GetCvMutex().unlock();

  // Computational vectors[NumProcThreads][MaxChannelsPerProcThread][this->FftLen]
  AVX_Vector<AVX_Vector<{VAL params/processing_data_type_samp}>>
      BUF_IN( _Obj.MaxChannelsPerProcThread,
      AVX_Vector<{VAL params/processing_data_type_samp}>(_Obj.FftLen[0],0) );  // [IN:len N] [Zeros: M-N]
  AVX_Vector<ComplexHermitionVector<{VAL params/processing_data_type_fd}>>
      BUF_A( _Obj.MaxChannelsPerProcThread,
      ComplexHermitionVector<{VAL params/processing_data_type_fd}>(_Obj.FftLen[0]) );  // [A:all]
  AVX_Vector<AVX_Vector<{VAL params/processing_data_type_samp}>>
      BUF_BC( _Obj.MaxChannelsPerProcThread,
      AVX_Vector<{VAL params/processing_data_type_samp}>(_Obj.FftLen[0],0) );  // [B:len N] [C:len M-N]
  AVX_Vector<AVX_Vector<{VAL params/processing_data_type_samp}>>
      BUF_OUT_OVLP( _Obj.MaxChannelsPerProcThread,
      AVX_Vector<{VAL params/processing_data_type_samp}>(_Obj.FftLen[0],0) );  // [Out;len N] [OVLP:len M-N]
  AVX_Vector<AVX_Vector<{VAL params/processing_data_type_samp}>>
      BUF_LAST_OVLP( _Obj.MaxChannelsPerProcThread,
      AVX_Vector<{VAL params/processing_data_type_samp}>(_Obj.FftLen[0],0) );
  AVX_Vector<AVX_Vector<{VAL params/processing_data_type_samp}>>
      BUF_OUTPUT( _Obj.MaxChannelsPerProcThread,
      AVX_Vector<{VAL params/processing_data_type_samp}>(_Obj.FftLen[0],0) );
  AVX_Vector<AVX_Vector<{VAL params/processing_data_type_fd}>>
      BUF_COEFS_Full( _Obj.MaxChannelsPerProcThread,
      AVX_Vector<{VAL params/processing_data_type_fd}>(_Obj.FftLen[0],0) );  // Local store for filter coefs
  AVX_Vector<ComplexHermitionVector<{VAL params/processing_data_type_fd}>>
      BUF_COEFS( _Obj.MaxChannelsPerProcThread,
      ComplexHermitionVector<{VAL params/processing_data_type_fd}>(_Obj.FftLen[0]) );  // Local store for filter coefs
  vector<vector<fftw_plan>>
      FFTW_Plan( _Obj.MaxChannelsPerProcThread, vector<fftw_plan>( 2, nullptr ));

  _Obj.FFTW_Mutex.lock();

  // Set-up FFT plans
  for( auto DataChan = LocChanBase ; DataChan < (LocChanBase+LocNumChans) ; ++DataChan )
  {
    lock_guard<mutex>  Lock(FFTW_Mut);

    ChanOffset = DataChan - LocChanBase;

    if( FFTW_Plan[ChanOffset][0] != nullptr )
    {
      fftw_destroy_plan( FFTW_Plan[ChanOffset][0] );
      FFTW_Plan[ChanOffset][0] = nullptr;
    }

    FFTW_Plan[ChanOffset][0] =  fftw_plan_dft_r2c_1d
                                (
                                  _Obj.FftLen[DataChan],
                                  &BUF_IN[ChanOffset][0],
                                  (fftw_complex*) &BUF_A[ChanOffset].AccessToFoldedVector()[0],
                                  FFTW_ESTIMATE
                                );

    assert(FFTW_Plan[ChanOffset][0]);

    if( FFTW_Plan[ChanOffset][1] != nullptr )
    {
      fftw_destroy_plan( FFTW_Plan[ChanOffset][1] );
      FFTW_Plan[ChanOffset][1] = nullptr;
    }

    FFTW_Plan[ChanOffset][1] =  fftw_plan_dft_c2r_1d
                                (
                                  _Obj.FftLen[DataChan],
                                  (fftw_complex*) &BUF_A[ChanOffset].AccessToFoldedVector()[0],
                                  &BUF_BC[ChanOffset][0],
                                  FFTW_ESTIMATE
                                );

    assert(FFTW_Plan[ChanOffset][1]);
  }

  _Obj.FFTW_Mutex.unlock();

  { // Signal we're alive and ready to process data
    unique_lock<mutex>   LK(_Obj.GetCvMutex());
    ++_Obj.Alive;
    LK.unlock();
    _Obj.GetCv().notify_all();
  }

  // Look for a Kill signal
  if( _Obj.KillFlag )  return RTIF::Exit_Due_To_Kill;

  // Wait for filter coefficient generation to finish before starting
  {
    unique_lock<mutex>  LK(_Obj.GetCvMutex());

    _Obj.GetCv().wait
        ( LK, [&]{ return (_Obj.StartUpState == _Obj.e_ReadFilterCoefs) || _Obj.KillFlag; } );

    // Look for a Kill signal
    if( _Obj.KillFlag )  return RTIF::Exit_Due_To_Kill;
  }

  // Set-up and cache the Write Schema
  const auto&  CoefSchema = _Obj.FRT_In.GetReadSchema_View();

  // Sync with Coef FIFO
  LocVect_CoefsIn[0] = CoefSchema.Start;

  // Copy the filter coefficients locally
  for( auto DataChan = LocChanBase ; DataChan < (LocChanBase+LocNumChans) ; ++DataChan )
  {
    ChanOffset = DataChan - LocChanBase;

    // Copy the filters down for use locally
    LocVect_CoefsIn[1] = DataChan;

    _Obj.FirFilter.WriteFifoToVector<{VAL params/processing_data_type_fd}>
    (
      LocVect_CoefsIn,
      LocVect_CoefsIn_Incr,
      &BUF_COEFS_Full[ChanOffset][0],
      _Obj.FftLen[0]
    );

    // Copy coefficient into ComplexHermitionVector
    BUF_COEFS[ChanOffset].ConvFullToFolded( BUF_COEFS_Full[ChanOffset] );
  }

  do
  {
    { // Wait for the cycle count to increment
      unique_lock<mutex>   LK(_Obj.GetCvMutex());
      _Obj.GetCv().wait( LK, [&]{ return (_Obj.CycleCounter != _Obj.ThreadCycleCounter[_ThreadNum]) ||
                                     _Obj.InData.GetEndOfDataFlag() || _Obj.KillFlag; } );
      ++_Obj.ThreadCycleCounter[_ThreadNum];
    }

    // Look for a Kill signal
    if( _Obj.KillFlag )  break;

    // Look for end of data
    if( _Obj.InData.GetEndOfDataFlag() )  break;

    ///////////////////////
    // Over data channels
    ///////////////////////

    // Set-up and cache the Schemas
    const auto&  ReadSchema  = _Obj.FRT_In.GetReadSchema_View();
    const auto&  WriteSchema = _Obj.FWT_Out.GetWriteSchema_View();

    // Sync with FIFOs
    LocVect_In[0]  = ReadSchema.Start;
    LocVect_Out[0] = WriteSchema.Start;

    for( auto DataChan = LocChanBase ; DataChan < (LocChanBase+LocNumChans) ; ++DataChan )
    {
      ChanOffset = DataChan - LocChanBase;

      // Sync with FIFOs
      LocVect_In[1]  = DataChan;
      LocVect_Out[1] = DataChan;

      _Obj.InData.WriteFifoToVector<double>
      (
        LocVect_In,
        LocVect_In_Incr,
        &BUF_IN[ChanOffset][0],
        ReadSchema.ActualReadLength
      );

      // Zero end of array
      for( uint32_t K = _Obj.FRT_In.GetActualReadLength() ;
                            K < _Obj.FRT_In.GetRequestedReadLength() ; ++K )
      {
        BUF_IN[ChanOffset][K] = 0;
      }

      // Perform FFT on 'BUF_IN[ChanOffset]' --> 'BUF_A[ChanOffset]'
      fftw_execute( FFTW_Plan[ChanOffset][0] );

      // Apply the Fourier Transform of the filter coefficients
      BUF_A[ChanOffset] *= BUF_COEFS[ChanOffset];

      // Perform IFFT on 'BUF_A[ChanOffset]' --> 'BUF_BC[ChanOffset]'
      fftw_execute( FFTW_Plan[ChanOffset][1] );

      // Apply FFT Scale
      const double  InvFactor = 1. / _Obj.FftLen[ChanOffset];

      for( auto K=0U ; K < _Obj.FftLen[ChanOffset] ; ++K )
      {
        BUF_BC[ChanOffset][K] *= InvFactor;
      }

      // Add the overlap
      for( auto K=0U ; K < _Obj.READ[ChanOffset] ; ++K )
      {
        BUF_BC[ChanOffset][K] += BUF_LAST_OVLP[ChanOffset][K];
      }

      // Assemble the output
      std::copy( &BUF_BC[ChanOffset][0],
                 &BUF_BC[ChanOffset][_Obj.READ[ChanOffset]],
                 &BUF_OUTPUT[ChanOffset][0] );

      // Save the new overlap
      std::copy( &BUF_BC[ChanOffset][_Obj.READ[DataChan]],
                 &BUF_BC[ChanOffset][_Obj.FftLen[DataChan]],
                 &BUF_LAST_OVLP[ChanOffset][0] );

      // Copy the data to 'FIFO'
      _Obj.{VAL params/output_fifo_label}.  WriteVectorToFifo
      (
        &BUF_OUTPUT[ChanOffset][0],
        LocVect_Out,
        LocVect_Out_Incr,
        WriteSchema.Length
      );
    }

    /////////////////////////////////
    // End of Processing
    /////////////////////////////////

    // Show that we are complete and ready to process another batch
    {
      unique_lock<mutex>   LK(_Obj.GetCvMutex());
      ++_Obj.Complete;
      LK.unlock();
      _Obj.GetCv().notify_all();
    }

  } while( !_Obj.KillFlag || !_Obj.InData.GetEndOfDataFlag() );

  // Clean-up FFT design info
  for( auto K=0U ; K < _Obj.MaxChannelsPerProcThread ; ++K )
  {
    lock_guard<mutex>  Lock(FFTW_Mut);

    if( FFTW_Plan[K][0] != nullptr )  fftw_destroy_plan(FFTW_Plan[K][0]);
    if( FFTW_Plan[K][1] != nullptr )  fftw_destroy_plan(FFTW_Plan[K][1]);
  }

  return  RTIF::Exit_Normally;
}


RTIF::ExitStateType  {VAL label}::KillThreads(void)
{
  this->State = State_Kill;

  { // Set Kill flags in all FIFOs
    {VAL params/output_fifo_label}.SetKillFlag(true);
    this->InData.SetKillFlag(true);
  }

  { // Set KillFlag in threads
    unique_lock<mutex>   LK(GetCvMutex());
    this->KillFlag = true;

    LK.unlock();

    GetCv().notify_all();
  }

  // Join up with the dying Control Thread
  ControlThread.join();

  // Now join up with the dying processing threads
  for( auto K=0U ; K < this->NumProcThreads ; ++K )
  {
    Thread[K].join();
  }

  if( { VAL debug } )
  { // Display Display_InitReset
    cout << Display_Kill();
  }

  return  RTIF::Exit_Due_To_Kill;
}


RTIF::ExitStateType  {VAL label}::Exit(void)
{
  this->State = State_Exit;

  { // Set Kill flags in all FIFOs
    {VAL params/output_fifo_label}.SetKillFlag(true);
    this->InData.SetKillFlag(true);
  }

  { // Set KillFlag in threads
    unique_lock<mutex>   LK(GetCvMutex());
    this->KillFlag = true;
    LK.unlock();
    GetCv().notify_all();
  }

  // Join up with the dying Control Thread
  ControlThread.join();

  // Now join up with the dying processing threads
  for( auto K=0U ; K < this->NumProcThreads ; ++K )
  {
    Thread[K].join();
  }

  if( { VAL debug } )
  { // Display Display_InitReset
    cout << Display_Exit();
  }

  return  RTIF::Exit_Normally;
}


///////////////////////////////////////
// Display and Debug Routines
///////////////////////////////////////

string  {VAL label}::Display_Constructor(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex());
  ostringstream       oS;

  oS << ">> {VAL label}::Constructor[" << this << "]" << endl;

  // Parameters
  oS << "  -NumProcThreads: " << this->NumProcThreads_FromParams << endl;
  oS << "  -NumChannels: " << this->NumChannels << endl;

  oS << "<< {VAL label}::Constructor[" << this << "]" << endl << endl;

  return  oS.str();
}


string  {VAL label}::Display_InitAlloc(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex());
  ostringstream       oS;

  oS << ">> {VAL label}::InitAlloc[" << this << "]" << endl;

  oS << "  -NumProcThreads: " << this->NumProcThreads << endl;
  oS << "  -NumChannels: " << this->NumChannels << endl;
  oS << "  -Channels:" << endl;
  oS << "  -FirFilterLength: " << this->FiltLen[0] << endl;
  oS << "  -FilterFftLen: " << this->FftLen[0] << endl;

  for( auto K=0U ; K < this->NumChannels ; ++K )
  {
    oS << "    -Channel #: " << K << endl;
    oS << "      -READ: " << this->READ[K] << endl;
    oS << "      -Lag: " << this->FilterStartUpLag[K] << endl;
  }

  oS << "<< {VAL label}::InitAlloc[" << this << "]" << endl << endl;

  return    oS.str();
}


string  {VAL label}::Display_InitReset(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex());
  ostringstream       oS;

  oS << ">> {VAL label}::InitReset[" << this << "]" << endl;

  oS << "  -NumProcThreads: " << this->NumProcThreads << endl;
  oS << "  -NumChannels: " << this->NumChannels << endl;
  oS << "  -Channels:" << endl;
  oS << "  -FirFilterLength: " << this->FiltLen[0] << endl;
  oS << "  -FilterFftLen: " << this->FftLen[0] << endl;

  for( auto K=0U ; K < this->NumChannels ; ++K )
  {
    oS << "    -Channel #: " << K << endl;
    oS << "      -READ: " << this->READ[K] << endl;
    oS << "      -Lag: " << this->FilterStartUpLag[K] << endl;
   }

  oS << "<< {VAL label}::InitReset[" << this << "]" << endl << endl;

  return    oS.str();
}


string  {VAL label}::Display_Kill(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex());
  ostringstream       oS;

  oS << ">> {VAL label}::Kill[" << this << "]" << endl;

  oS << "  -KillFlag: " << boolalpha << this->KillFlag << endl;
  oS << "  -CycleCounter: " << this->CycleCounter << endl;
  oS << "  -Complete: " << this->Complete << endl;

  for( auto K=0U ; K < {VAL params/num_proc_threads} ; ++K )
  {
    oS << "    -Thread[" << setw(2) << K << "]:" << endl;
   }

  oS << "<< {VAL label}::Kill[" << this << "]" << endl << endl;

  return    oS.str();
}


string  {VAL label}::Display_Exit(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex());
  ostringstream       oS;

  oS << ">> {VAL label}::Exit[" << this << "]" << endl;

  oS << "  -KillFlag: " << boolalpha << this->KillFlag << endl;
  oS << "  -CycleCounter: " << this->CycleCounter << endl;
  oS << "  -Complete: " << this->Complete << endl;

  for( auto K=0U ; K < {VAL params/num_proc_threads} ; ++K )
  {
    oS << "    -Thread[" << setw(2) << K << "]:" << endl;
   }

  oS << "<< {VAL label}::Exit[" << this << "]" << endl << endl;

  return    oS.str();
}


