///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  SpectrumGenerator.cpp                                                             //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//
//<LIBTAG:c029de4c-a6bc-41dc-88d0-3222c8907d78>

#include  "modules.h"
#include  "SigProcWindows.h"
#include  "Exception.h"
#include  "{VAL label}.h"


// Spectral window object
SigProcWindows<double,{VAL params/fft_len}>   SigProcWindowsObj;


{VAL label}::{VAL label}(void) :
StartUpState(e_InitState),
InData({VAL params/input_rfifo}),
FRT_In(InData),
PostFilt_Dimensions  {
                       {VAL params/output_fifo_length},
                       {VAL params/fft_len},
                       {VAL params/num_channels}
                     },
PostFilt_WrapFlags  { true, false, false },
PostFilt_Extent  { PostFilt_Dimensions, PostFilt_WrapFlags },
{VAL params/output_fifo_label}(PostFilt_Extent),
FWT_Out({VAL params/output_fifo_label})
{
  this->State = State_Initialized;

  if( { VAL debug } )
  { // Debug display Display_Constructor
    cout << this->Display_Constructor();
  }
}


bool  {VAL label}::AnyActiveThreads(void) const
{
  bool  AnyActiveThreads = false;

  AnyActiveThreads = AnyActiveThreads || this->ControlThread.joinable();

  for( auto K=0U ; K < this->NumProcThreads ; ++K )
  {
    AnyActiveThreads = AnyActiveThreads || this->Thread[K].joinable();
  }

  return  AnyActiveThreads;
}


RTIF::ExitStateType  {VAL label}::Init_Allocate(void)
{
  this->State = State_Allocated;

  // Calculate 'NumProcThreads'
  if( this->NumProcThreads_FromParams <= this->NumChannels )
  {
    this->NumProcThreads =  this->NumProcThreads_FromParams;
  }
  else
  { // No need to spawn more threads than channels
    this->NumProcThreads = this->NumChannels;
  }

  // We need this value for allocating maximum array sizes during processing
  this->MaxChannelsPerProcThread = 0;

  // Determine which data channels are processed for each thread
  for( uint32_t K=0U ; K < this->NumProcThreads ; ++K )
  {
    this->BaseChannelNumberForThread
    (
      K,
      this->NumProcThreads,
      this->NumChannels,
      this->ChannelStart[K],
      this->NumChannelsPerThread[K]
    );

    if( this->NumChannelsPerThread[K] > this->MaxChannelsPerProcThread )
    {
      this->MaxChannelsPerProcThread = this->NumChannelsPerThread[K];
    }
  }

  // If there are any currently active threads kill them and restart
  if( this->AnyActiveThreads() )
  { // If there are active threads: kill them
    this->KillThreads();
    this->KillFlag = false;
    this->State = State_Allocated;
  }

  // Reset some control variables
  this->Alive         = 0;
  this->Complete      = 0;
  this->CycleCounter  = ~0;
  this->StartUpState  = this->e_InitState;

  if( { VAL debug } )
  { // Display Display_InitAlloc
    cout << Display_InitAlloc();
  }

  return  RTIF::Exit_Normally;
}


RTIF::ExitStateType  {VAL label}::Init_Reset(void)
{
  this->State = State_Reset;

  // If there are any currently active threads kill them and restart
  if( this->AnyActiveThreads() )
  { // There are active threads: kill them
    this->KillThreads();
    this->KillFlag = false;
    this->State = State_Reset;
  }

  // Reset some control variables
  this->Alive         = 0;
  this->Complete      = 0;
  this->CycleCounter  = ~0;
  this->StartUpState  = this->e_InitState;

  // Spawn the controller thread
  this->ControlThread = thread( this->ProcessBatches, ref(*this) );

  // Spawn the processing threads
  for( uint32_t K=0U ; K < this->NumProcThreads ; ++K )
  {
    this->ThreadCycleCounter[K] = ~0;
    this->Thread[K]             = thread( this->ProcessingThread, ref(*this), K );
  }

  { // Wait for all the threads to wake up and then wait on CycleCounter
    unique_lock<mutex>    LK(this->GetCvMutex);
    this->CV.wait( LK, [&]{ return (this->Alive == this->NumProcThreads) || this->KillFlag; } );
  }

  if( { VAL debug } )
  { // Display Display_InitReset
    cout << Display_InitReset();
  }

  return  RTIF::Exit_Normally;
}


// Assuming all filters have the same number of coefficients
RTIF::ExitStateType  {VAL label}::ProcessBatches( {VAL label}& _Obj )
{
  // Look for a Kill signal
  if( _Obj.KillFlag )  return  RTIF::Exit_Due_To_Kill;

  /////////////////////////
  // Main Processing Loop
  /////////////////////////

  do
  {
    // Wait for input data
    _Obj.InData.FIFO_Read_Fixed_Amount( _Obj.FftLen, _Obj.FRT_In );

    // Wait for output space
    _Obj.{VAL params/output_fifo_label}.FIFO_Write_Start( 1, _Obj.FWT_Out );

    // Look for a Kill signal
    if( _Obj.KillFlag )  break;

    { // Increment the cycle count to start processing
      unique_lock<mutex>   LK(_Obj.GetCvMutex);
      ++_Obj.CycleCounter;
      _Obj.Complete = 0;
      LK.unlock();
      _Obj.CV.notify_all();
    }

    // Look for a Kill signal
    if( _Obj.KillFlag )  break;

    /////////////////////////////////
    // Processing occurs here
    /////////////////////////////////

    { // Wait for children to finish
      unique_lock<mutex>   LK(_Obj.GetCvMutex);
      _Obj.CV.wait( LK, [&]{ return (_Obj.Complete == _Obj.NumProcThreads) || _Obj.KillFlag; } );
    }

    // Look for a Kill signal
    if( _Obj.KillFlag )  break;

    // Done writing into FIFO
    _Obj.{VAL params/output_fifo_label}.FIFO_Write_End( _Obj.FWT_Out );

    // Look for a Kill signal
    if( _Obj.KillFlag )  break;

    // Done reading from FIFO
    _Obj.InData.FIFO_Read_End( _Obj.FRT_In, _Obj.Advance );

  }  while(!_Obj.KillFlag);

  return  RTIF::Exit_Due_To_Kill;
}


RTIF::ExitStateType  {VAL label}::ProcessingThread( {VAL label}& _Obj, const uint32_t _ThreadNum )
{
  uint32_t                ChanOffset = 0U;
  vector<uint32_t>        LocVect_In(_Obj.InData.GetVectorLen());
  vector<uint32_t>        LocVect_Out(_Obj.{VAL params/output_fifo_label}.GetVectorLen());
  const vector<uint32_t>  LocVect_In_Incr = { 1, 0 };
  const vector<uint32_t>  LocVect_Out_Incr = { 0, 0, 1 };

  // Grab local processing parameters
  _Obj.GetCvMutex.lock();
  const auto    LocChanBase   = _Obj.ChannelStart[_ThreadNum];
  const auto    LocNumChans   = _Obj.NumChannelsPerThread[_ThreadNum];
  _Obj.GetCvMutex.unlock();

  // Computational vectors[NumProcThreads][MaxChannelsPerProcThread][this->FftLen]
  AVX_Vector<AVX_Vector<{VAL params/proc_t_type}>>
      BUF_IN_D( _Obj.MaxChannelsPerProcThread,
      AVX_Vector<{VAL params/proc_t_type}>(_Obj.FftLen,0) );
  AVX_Vector<ComplexHermitionVector<complex<double>>>
      BUF_OUT_D( _Obj.MaxChannelsPerProcThread,
      ComplexHermitionVector<complex<double>>(_Obj.FftLen) );
  AVX_Vector<AVX_Vector<{VAL params/proc_m_type}>>
      BUF_OUT_I( _Obj.MaxChannelsPerProcThread,
      AVX_Vector<{VAL params/proc_m_type}>(_Obj.FftLen,0) );
  AVX_Vector<AVX_Vector<{VAL params/proc_m_type}>>
      BUF_WINDOW_D( _Obj.MaxChannelsPerProcThread,
      AVX_Vector<{VAL params/proc_m_type}>(_Obj.FftLen) );
  vector<fftw_plan>  FFTW_Plan( _Obj.MaxChannelsPerProcThread, nullptr );

  // Set-up FFT plans
  for( auto DataChan = LocChanBase ; DataChan < (LocChanBase+LocNumChans) ; ++DataChan )
  {
    lock_guard<mutex>  Lock(FFTW_Mut);

    ChanOffset = DataChan - LocChanBase;

    if( FFTW_Plan[ChanOffset] != nullptr )
    {
      fftw_destroy_plan( FFTW_Plan[ChanOffset] );
      FFTW_Plan[ChanOffset] = nullptr;
    }

    FFTW_Plan[ChanOffset] =  fftw_plan_dft_r2c_1d
                             (
                               _Obj.FftLen,
                               &BUF_IN_D[ChanOffset][0],
                               (fftw_complex*) &BUF_OUT_D[ChanOffset].AccessToFoldedVector()[0],
                               FFTW_ESTIMATE
                             );

    assert(FFTW_Plan[ChanOffset]);
  }

  // Load the appropriate spectral window coefficients
  BUF_WINDOW_D[ChanOffset] = SigProcWindowsObj.GetSpectralWindow("BlackmanHarris_92");

  { // Signal we're alive and ready to process data
    unique_lock<mutex>   LK(_Obj.GetCvMutex);
    ++_Obj.Alive;
    LK.unlock();
    _Obj.CV.notify_all();
  }

  // Look for a Kill signal
  if( _Obj.KillFlag )  return ExitStateType::Exit_Due_To_Kill;

  do
  {
    { // Wait for the cycle count to increment
      unique_lock<mutex>   LK(_Obj.GetCvMutex);
      _Obj.CV.wait( LK, [&]{ return (_Obj.CycleCounter != _Obj.ThreadCycleCounter[_ThreadNum]) || _Obj.KillFlag; } );
      ++_Obj.ThreadCycleCounter[_ThreadNum];
    }

    // Look for a Kill signal
    if( _Obj.KillFlag )  break;

    // Over data channels
    for( auto DataChan = LocChanBase ; DataChan < (LocChanBase+LocNumChans) ; ++DataChan )
    {
      ChanOffset = DataChan - LocChanBase;

      // Copy the data from 'FIFO_Reader'
      LocVect_In[1] = DataChan;

      _Obj.InData.WriteFifoToVector<{VAL params/proc_t_type}>
      (
        LocVect_In,
        LocVect_In_Incr,
        &BUF_IN_D[ChanOffset][0],
        {VAL params/fft_len}
      );

      // Set-up and cache the Schemas
      const auto&  WriteSchema = _Obj.FWT_Out.GetWriteSchema_View();

      // Apply the spectral window
      for( auto K=0U ; K < {VAL params/fft_len} ; ++K )
      {
        BUF_IN_D[ChanOffset][K] = BUF_IN_D[ChanOffset][K] * BUF_WINDOW_D[ChanOffset][K];
      }

      // Perform FFT on 'BUF_IN_D[ChanOffset]' --> 'BUF_OUT_D[ChanOffset]'
      fftw_execute( FFTW_Plan[ChanOffset] );

      // Recover spectral aplitudes
      for( auto K=0U ; K < _Obj.FftLen ; ++K )
      {
        BUF_OUT_I[ChanOffset][K] = std::abs( BUF_OUT_D[ChanOffset][K] );
      }

      /////////////////////////////////////////////////
      // Write the data into the output FIFO
      /////////////////////////////////////////////////

      LocVect_Out[0] = WriteSchema.Start;
      LocVect_Out[1] = DataChan;
      LocVect_Out[2] = 0;

      _Obj.{VAL params/output_fifo_label}.WriteVectorToFifo<{VAL params/proc_m_type}>
      (
        &BUF_OUT_I[ChanOffset][0],
        LocVect_Out,
        LocVect_Out_Incr,
        _Obj.FftLen
      );
    }

    /////////////////////////////////
    // End of Processing
    /////////////////////////////////

    // Show that we are complete and ready to process another batch
    {
      unique_lock<mutex>   LK(_Obj.GetCvMutex);
      ++_Obj.Complete;
      LK.unlock();
      _Obj.CV.notify_all();
    }

  } while(!_Obj.KillFlag);

  // KILL FLAG

  for( auto K=0U ; K < _Obj.MaxChannelsPerProcThread ; ++K )
  {
    if( FFTW_Plan[K] != nullptr )
    {
      lock_guard<mutex>  Lock(FFTW_Mut);

      fftw_destroy_plan(FFTW_Plan[K]);
    }
  }

  return  RTIF::Exit_Due_To_Kill;
}


RTIF::ExitStateType  {VAL label}::KillThreads(void)
{
  this->State = State_Kill;

  { // Set Kill flags in all FIFOs
    {VAL params/output_fifo_label}.SetKillFlag(true);
    this->InData.SetKillFlag(true);
  }

  { // Set KillFlag in threads
    unique_lock<mutex>   LK(GetCvMutex);
    this->KillFlag = true;

    LK.unlock();

    CV.notify_all();
  }

  // Join up with the dying Control Thread
  ControlThread.join();

  // Now join up with the dying processing threads
  for( auto K=0U ; K < this->NumProcThreads ; ++K )
  {
    Thread[K].join();
  }

  if( { VAL debug } )
  { // Display Display_InitReset
    cout << Display_Kill();
  }

  return  RTIF::Exit_Normally;
}


///////////////////////////////////////
// Display and Debug Routines
///////////////////////////////////////

string  {VAL label}::Display_Constructor(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex);
  ostringstream       oS;

  oS << ">> {VAL label}::Constructor[" << this << "]" << endl;

  // Parameters
  oS << "  -NumProcThreads: " << this->NumProcThreads_FromParams << endl;
  oS << "  -NumChannels: " << this->NumChannels << endl;

  oS << "<< {VAL label}::Constructor[" << this << "]" << endl << endl;

  return  oS.str();
}


string  {VAL label}::Display_InitAlloc(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex);
  ostringstream       oS;

  oS << ">> {VAL label}::InitAlloc[" << this << "]" << endl;

  oS << "  -NumProcThreads: " << this->NumProcThreads << endl;
  oS << "  -NumChannels: " << this->NumChannels << endl;
  oS << "  -Channels:" << endl;

  for( auto K=0U ; K < this->NumChannels ; ++K )
  {
    oS << "    -Channel #: " << K << endl;
  }

  oS << "<< {VAL label}::InitAlloc[" << this << "]" << endl << endl;

  return    oS.str();
}


string  {VAL label}::Display_InitReset(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex);
  ostringstream       oS;

  oS << ">> {VAL label}::InitReset[" << this << "]" << endl;

  oS << "  -NumProcThreads: " << this->NumProcThreads << endl;
  oS << "  -NumChannels: " << this->NumChannels << endl;
  oS << "  -Channels:" << endl;

  for( auto K=0U ; K < this->NumChannels ; ++K )
  {
    oS << "    -Channel #: " << K << endl;
   }

  oS << "<< {VAL label}::InitReset[" << this << "]" << endl << endl;

  return    oS.str();
}


string  {VAL label}::Display_Kill(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex);
  ostringstream       oS;

  oS << ">> {VAL label}::Kill[" << this << "]" << endl;

  oS << "  -KillFlag: " << boolalpha << this->KillFlag << endl;
  oS << "  -CycleCounter: " << this->CycleCounter << endl;
  oS << "  -Complete: " << this->Complete << endl;

  for( auto K=0U ; K < {VAL params/num_proc_threads} ; ++K )
  {
    oS << "    -Thread[" << setw(2) << K << "]:"
       << endl;
   }

  oS << "<< {VAL label}::Kill[" << this << "]" << endl << endl;

  return    oS.str();
}


