///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  InverseParseFrame.h                                                               //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#pragma once

#include    <cassert>
#include    <iostream>
#include    <list>
#include    <vector>
#include    <string>
#include    "Outputs.h"

using namespace std;


namespace   wej_parser
{

class     GrammarAbs;

///////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////
// class    InverseParseFrame
///////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////

class   InverseParseFrame
{
private:

  // Parameters
  ostream&              Out;
  Outputs&              List;
  bool                  DebugOutput;
  InverseParseFrame*    Father;
  GrammarAbs*           Rule;

  // This flag is sent on the first IPF.  When a TAG object is inverse parsed the list
  //   position is incremented first UNLESS this flag is true.  It is set FALSE on execution
  //   of the first TAG inverse parse.
  bool  ListInInit;

  // Current List position
  list<OutputRec>::const_iterator   CurrListPos;

  // Output file position
  streampos   FileP0;            // On Entry
  streampos   FileP;             // On Mark

  // List Positions
  list<OutputRec>::const_iterator   ListP0;
  list<OutputRec>::const_iterator   ListP;

  // Marking
  inline void    MarkFile(void)
  {
    this->FileP = this->GetOutPos();
  }

  inline void    MarkList(void)
  {
    this->ListP = this->CurrListPos;
  }

  inline void    Mark(void)
  {
    this->MarkFile();
    this->MarkList();
  }

  // Used in initialization so we can use Mark
  inline void    SetFirstFromMark(void)
  {
    this->FileP0  = this->FileP;
    this->ListP0  = this->ListP;
  }

  // Setting Position
  inline void    MoveFileTo( const streampos  NewPos )
  {
    this->Out.clear();
    this->Out.seekp( NewPos, ios::beg );
  }

  inline void   MoveListTo( const list<OutputRec>::const_iterator NewPos )
  {
    this->CurrListPos   = NewPos;
  }

public:

  inline  InverseParseFrame( Outputs& pInList, ostream& pOut, const bool pDebugOutput=false ) :
                             Out(pOut), List(pInList), DebugOutput(pDebugOutput), Father(nullptr),
                             Rule(nullptr), ListInInit(true), CurrListPos(List.Begin())
  {
    this->Mark();
    this->SetFirstFromMark();
  }

  inline  InverseParseFrame( InverseParseFrame& Other ) :
                             Out(Other.Out), List(Other.List), DebugOutput(Other.DebugOutput),
                             Father(&Other), Rule(nullptr), ListInInit(Other.ListInInit),
                             CurrListPos(Other.CurrListPos)
  {
    this->Mark();
    this->SetFirstFromMark();
  }

  ~InverseParseFrame(void);

  inline void       SetRule( GrammarAbs* pRule )
  {
    this->Rule  = pRule;
  }

  inline GrammarAbs*    GetRule(void)
  {
    return  this->Rule;
  }

  // Position Capture
  inline streampos    GetOutPos(void) const
  {
    return  this->Out.tellp();
  }

  inline bool       IsDebug(void) const
  {
    return  this->DebugOutput;
  }

  inline void       AcceptAll(void)
  {
    this->Mark();
  }

  inline void       AcceptOut(void)
  {
    this->MarkFile();
  }

  inline void       AcceptList(void)
  {
    this->MarkList();
  }

  inline bool       IsDebugOutput(void) const
  {
    return  this->DebugOutput;
  }

  inline list<OutputRec>::const_iterator     ListBeg(void) const
  {
    return  this->List.Begin();
  }

  inline list<OutputRec>::const_iterator     ListEnd(void) const
  {
    return  this->List.End();
  }

  inline bool   GetListInInit(void) const
  {
    return  this->ListInInit;
  }

  inline void   SetListInInit( const bool SetTo )
  {
    this->ListInInit = SetTo;
  }

  inline void   IncrementList(void)
  {
    if( this->CurrListPos != this->List.End() )   ++this->CurrListPos;
  }

  inline list<OutputRec>::const_iterator  GetListPos(void) const
  {
    return  this->CurrListPos;
  }

  inline Outputs&     GetOutputs(void)
  {
    return  this->List;
  }

  inline  void  WriteStringToOutput( const string& S )
  {
    Out.write( S.c_str(), S.length() );
  }

  inline  bool  IsInEof(void) const
  {
    return  this->CurrListPos == this->List.End();
  }

  // Display
  string    Display(void) const;
};
}   // End namespace 'wej_parser'


