///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  build_manifest.cpp                                                                //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#include  <iostream>
#include  <fstream>
#include  <filesystem>
#include  <string>
#include  <vector>
#include  <iomanip>
#include  <map>
#include  "config.h"
#include  "Translator.h"
#include  "XML_Storage.h"
#include  "XML_Node.h"

namespace fs = std::filesystem;

using namespace std;

// Root file folder
extern string  PES_Tools_Root;

////////////////////////////////////////////////////////////////////////////////////////////////
// PES Tools Directory Structure.  Creates: 'lib_manifest.xml'
////////////////////////////////////////////////////////////////////////////////////////////////
//
//  [PES_TOOLS]
//    lib_manifest.xml
//    [tools]
//      [build_manifest]
//      [xml_decorator]
//    [lib]
//      [any set of subdirectories]
//        [any type of source file containing a UUID TAG such as:
//          <LIBTAG:9ea64c89-5b01-4aee-bd5d-2c564dd504aa>.
//          Note that the UUID tag is generated with 'uuidgen' on Linux.]
//
////////////////////////////////////////////////////////////////////////////////////////////////

string	RemoveHyphens( const string& Source )
{
  string  Ret;

  for( auto K=0U ; K < Source.size() ; ++K )
  {
  if( Source[K] == '-' )  continue;

  Ret += tolower(Source[K]);
  }
  return  Ret;
}


/////////////////////////////////////////////////////////////////////////////
// Search sub-directories recursively and look for LIBTAGS and return UUID.
//   e.g.  <LIBTAG:9ea64c89-5b01-4aee-bd5d-2c564dd504aa>
/////////////////////////////////////////////////////////////////////////////

int main( [[maybe_unused]] int NumArgs, [[maybe_unused]] char *ArgList[] )
{
  const bool  Debug = false;

  uint32_t    NumFiles = 0U;
  uint32_t    NumErrors = 0U;

  // Get timestamp
  std::time_t  NowTime = std::chrono::system_clock::to_time_t(std::chrono::system_clock::now());
  std::string  TimeString(16,' ');
  std::strftime(&TimeString[0], 16U, "%Y%m%d_%H%M%S", std::localtime(&NowTime));
  TimeString.resize(15);     // Fix trailing zero

  // Get paths with respect to the PES_TOOLS root
  fs::path  CurrPath = fs::current_path().lexically_normal();
  fs::path  RootPath = fs::current_path().append("../../").lexically_normal();
  fs::path  LibPath = RootPath;
  fs::path  ToolsPath = RootPath;
  fs::path  BuildManifestPath_cpp = RootPath;
  fs::path  XmlDecoratorPath = RootPath;

  LibPath.append("lib/").lexically_normal();
  ToolsPath.append("tools/").lexically_normal();
  BuildManifestPath_cpp.append("tools/build_manifest/pes_tools_path.cpp").lexically_normal();
  XmlDecoratorPath.append("tools/xml_decorator/pes_tools_path.cpp").lexically_normal();

  fs::path  XML_FilePath = RootPath;
  XML_FilePath.append("lib_manifest.xml").lexically_normal();

  if(Debug)
  {
    cout << "[+]build_manifest" << endl;
    cout << "   TimeString: " << TimeString << endl;
    cout << "   RootPath: " << RootPath.native() << endl;
    cout << "   LibPath: " << LibPath.native() << endl;
    cout << "   ToolsPath: " << ToolsPath.native() << endl;
    cout << "   BuildManifestPath: " << BuildManifestPath_cpp.native() << endl;
    cout << "   XmlDecoatorPath: " << XmlDecoratorPath.native() << endl;
    cout << "   XML_FilePath: " << XML_FilePath.native() << endl;
    cout << endl;
  }

  //////////////////////////////
  // Create path files to root
  //////////////////////////////

  {
    ofstream  Out( BuildManifestPath_cpp, ios::out | ios::trunc | ios::binary );

    Out << "#include  <filesystem>" << endl << endl;
    Out << "std::filesystem::path  PES_Tools_Root{ \"" << RootPath.native() << "\" };" << endl;
  }

  {
    ofstream  Out( XmlDecoratorPath, ios::out | ios::trunc | ios::binary );

    Out << "#include  <filesystem>" << endl << endl;
    Out << "std::filesystem::path  PES_Tools_Root{ \"" << RootPath.native() << "\" };" << endl;
  }

  //////////////////////////////
  // Create path files to root
  //////////////////////////////

  // We must be running in the correct directory for this program to work correctly
  string  LastVal;

  for( auto I=CurrPath.begin() ; I != CurrPath.end() ; ++I )
  {
    LastVal = *I;
  }

  if( LastVal != "build_manifest" )
  {
    cout << "*Error: This program can only be run in the 'build_manifest' subdirectory." << endl
         << "        Program Terminated..." << endl;

    return  1;
  }

  multimap<string,string> Map;		 // UUID <---> File Path
  list<string>            UUIDs;

  if(true)
  {
    cout << ">>build_manifest: Rebuilding \'lib_manifest.xml\' file" << endl;

    cout << "-Scanning Libraries for UUID tags, e.g."
         << " //<LIBTAG:756033ec-8d0a-4bac-909b-dcc4eba3680a>"
         << endl;

    cout << "-Root Path " << RootPath << endl;
  }

  // Scan all library files
  for( const fs::directory_entry& Entry : fs::recursive_directory_iterator(LibPath) )
  {
    const string            GrammarFileName = "find_uuid.bnf0";
    ifstream                In( Entry.path() );
    ifstream                Grammar(GrammarFileName);
    wej_parser::Outputs	    OutList;
    wej_parser::Translator  T(Grammar);

    fs::path  CurrPath, LibPath;

    // See if this is a directory
    fs::directory_entry  TestPath(Entry);

    if(false)
    {
      cout << endl;
      cout << "-" << TestPath.path() << endl;
      cout << "---is_directory:  " << boolalpha << TestPath.is_directory() << endl;
    }

    if( TestPath.is_directory() )  continue;

    if(true)
    {
      cout << "--" << TestPath.path() << endl;
    }

    // Run the grammar engine (Translation Grammar: In ---> OutList)
    const auto TransRet = T.Translate( In, OutList );

    // Error checking
    if( !T.IsGrammarGood() )
    {
      cout << ">>Error: Grammar cannot be compiled." << endl;
      return  1;
    }

    if( !TransRet )
    {
      cout << ">>Error: File \'" << Entry.path().native() <<  "\'" << endl;
      return  1;
    }

    // We must find a TAG
    if( OutList.NumRecords() == 0U )
    {
      cout << "*Error: Cannot find Library Tag in file: " << Entry.path() << endl;
      continue;
    }

    if( OutList.NumRecords() > 1U )
    {
      cout << "*Error: Found multiple UUID tags in file: " << Entry.path() << endl;
      continue;
    }

    // Get UUID
    bool    Found = false;
    string  UUID;

    for( auto L=OutList.Begin() ; (L != OutList.End()) && (!Found) ; ++L )  // Tag entries
    {
      UUID = L->GetDataVect()[1];
    }

    // Update the Map
    Map.insert( pair( UUID, Entry.path() ) );

    // Form a list of unique keys
    bool  FoundIt = false;

    for( auto I: UUIDs )
    {
      if( I != UUID )  continue;

      FoundIt = true;
      break;
    }

    if(!FoundIt)  UUIDs.push_back(UUID);

    ++NumFiles;
  }

  // Sort Keys
  UUIDs.sort();

  ////////////////////
  // XML output file
  ////////////////////

  wej_parser::XML_Storage  XML;

  ////////////////////
  // uuids
  ////////////////////

  // Save the root_path
  auto FileNameXML = wej_parser::XML_Node(nullptr);
  FileNameXML.SetLabel("root_dir");
  FileNameXML.SetData(RootPath);
  XML.GetHead().PushChild(FileNameXML);

  // Date/Time stamp
  auto DateTimeStamp = wej_parser::XML_Node(nullptr);
  DateTimeStamp.SetLabel("gen_date");
  DateTimeStamp.SetData(TimeString);
  XML.GetHead().PushChild(DateTimeStamp);

  // Create a directory to hold the library files indexed by their UUIDs
  auto Library = wej_parser::XML_Node(nullptr);
  Library.SetLabel("lib");

  auto  TotalCounter = 0U;
  auto  AutoRecordCounter = 0U;
  auto  FileCounter = 0U;

  for( auto I: UUIDs )
  { // uuids
    const string UuidTag = "u_" + RemoveHyphens(I);

    auto UuidItem = wej_parser::XML_Node(nullptr);
    UuidItem.SetLabel(UuidTag);

    auto FileList = wej_parser::XML_Node(nullptr);
    FileList.SetLabel("files");

    FileCounter = 0U;

    for( auto It = Map.find(I) ; (It != Map.end()) && (It->first == I) ; ++It )
    {
      // See if this a file that needs to be init'ed
      fs::path  Path = It->second;

      // Root-relative path
      Path = Path.lexically_relative(RootPath);

      // See if this file needs to be 'Init'ed
      const string  Extension = Path.extension();

      const bool  MayBeInited = (Extension.find("cpp") != string::npos) ||
                                (Extension.find("cc")  != string::npos) ||
                                (Extension.find("CPP") != string::npos) ||
                                (Extension.find("CC")  != string::npos)  ;

      auto FileEntry = wej_parser::XML_Node(nullptr);
      FileEntry.SetLabel("f_" + std::to_string(FileCounter++) );

      auto InitFlag = wej_parser::XML_Node(nullptr);    // Usually a compilable file such as "*.cpp"
      InitFlag.SetLabel("init");
      InitFlag.SetData(MayBeInited);
      FileEntry.PushChild(InitFlag);

      // File 'path'
      auto FileName = wej_parser::XML_Node(nullptr);
      FileName.SetLabel("path");

      // Remove the leading 'lib/' from path name
      fs::path  PathWithoutLibPrefix;

      fs::path::iterator  I;

      for( auto I = Path.begin() ; I != Path.end() ; ++I )
      {
        if( *I == "lib" )  continue;
        PathWithoutLibPrefix = PathWithoutLibPrefix / *I;
      }

      FileName.SetData(PathWithoutLibPrefix);
      FileEntry.PushChild(FileName);
      FileList.PushChild(FileEntry);
    }

    // Push Upwards
    UuidItem.PushChild(FileList);
    Library.PushChild(UuidItem);

    ++FileCounter;
    ++AutoRecordCounter;
    ++TotalCounter;
  }

  XML.GetHead().PushChild(Library);

  /////////////////////////////////////////
  // Root the tree onto the XML file tree
  /////////////////////////////////////////

  XML.GetHead().SetLabel("uuids");

  // Write output to XML file
  const auto Flag = XML.WriteToXML_File(XML_FilePath);
  assert(Flag);

  if(true)
  {
    cout << "-Created File '" << PES_Tools_Root << "lib_manifest.xml' with "
         << TotalCounter << " records" << endl;
    cout << "<<build_manifest: " << NumFiles << " files processed    Num Errors: " << NumErrors << endl;
  }

  return 0;
}


