///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  FIFO_Reader.h                                                                     //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#pragma once

#include  <iomanip>
#include  "FIFO.h"

using namespace std;

////////////////////////////////////////////////////////////////////////////////////////////////////
//  FIFO Addressing
////////////////////////////////////////////////////////////////////////////////////////////////////
//
// - FIFOs are defined by two N-Dimensional vectors:
//   - Extents (size of FIFO), e.g. [ 1024  2  128 ].
//   - Wrap flags: if TRUE the dimension wraps about the end point, e.g. [ true  false  false ].
// - FIFO locations are uniquely defined by a N-dimensional vector, e.g. [ X Y Z ].
// - Entries in a FIFO are memory contiguous over their rightmost entry (then moving leftward):
//   - e.g. [ 0 0 0 ], [ 0 0 1 ], ..., [ 0 1 0 ], [ 0 1 1 ], ..., [ 1 0 0 ], [ 1 0 1 ].
// - The left-most entry is special and is ALWAYS the FIFO Sample Number (FIFO column number):
//   - e.g. [ FifoSampleNumber ... ]
// - Other dimensions are at the discretion of the programmer and defined by the Extents vector.
//
////////////////////////////////////////////////////////////////////////////////////////////////////


// For debug prints
extern  mutex  DebugMutex;     // Global Debug Mutex


template<typename DataType,typename AllocType>
class  FifoReaderTool;


///////////////////////////////////////////////////////////////////////////
//  FIFO_Reader<typename DataType,typename AllocType>
///////////////////////////////////////////////////////////////////////////

template<typename DataType,typename AllocType>
class  FIFO_Reader
{
public:

  const bool  AssertDebugs = false;

  class  ReadLockTool
  {
  public:

    typedef struct sReadSchema
    {
      uint32_t    RequestedReadLength;
      uint32_t    ActualReadLength;
      uint32_t    Start;
      uint32_t    End;

      // Note that StartWrapped is assumed 0
      uint32_t    EndWrapped;
      bool        WrapFlag;

    }  ReadSchemaType;

  private:

    // Inputs
    FIFO_Reader<DataType,AllocType>&  FifoReaderRef;

    uint32_t  RequestedReadLength;
    uint32_t  ActualReadLength;

  public:

    ReadLockTool( FIFO_Reader<DataType,AllocType>& _FIFO_Reader ) :
    FifoReaderRef(_FIFO_Reader),
    RequestedReadLength(0U),
    ActualReadLength(0U)
    {
    }

    inline void  SetRequestedReadLength( const int32_t RequestedReadLengthParam )
    {
      this->RequestedReadLength = RequestedReadLengthParam;
    }

    inline uint32_t  GetRequestedReadLength(void) const
    {
      return  this->RequestedReadLength;
    }

    inline void  SetActualReadLength( const int32_t ActualReadLengthParam )
    {
      this->ActualReadLength = ActualReadLengthParam;
    }

    inline uint32_t  GetActualReadLength(void) const
    {
      return  this->ActualReadLength;
    }

    bool  IsFifoWrap(void) const
    {
      return  (this->FifoReaderRef.Back + this->NumData) >
                    this->FifoReaderRef.FifoRef.Extent.LengthLowestDim;
    }

    void  GetReadSchema( ReadSchemaType& _ReadSchema ) const
    {
      _ReadSchema.RequestedReadLength   = this->RequestedReadLength;
      _ReadSchema.ActualReadLength      = this->ActualReadLength;
      _ReadSchema.Start                 = this->FifoReaderRef.Back;

      const uint32_t  Top = _ReadSchema.Start + this->ActualReadLength;

      _ReadSchema.WrapFlag  = Top > this->FifoReaderRef.GetBufMap().GetLengthLowestDim();

      if( _ReadSchema.WrapFlag )
      { // Wrap around end of FIFO
        _ReadSchema.End           = this->FifoReaderRef.GetBufMap().GetLengthLowestDim();
        _ReadSchema.EndWrapped    = Top - this->FifoReaderRef.GetBufMap().GetLengthLowestDim();
      }
      else
      { // Single segment
        _ReadSchema.End           = _ReadSchema.Start + this->ActualReadLength;
        _ReadSchema.EndWrapped    = 0;
      }
    }

    /////////////////////////////
    // Physical read locations
    /////////////////////////////

    DataType*  Loc_Base0( const vector<uint32_t>& _LocationVector, const ReadSchemaType& _RTL ) const
    {
      return  &this->FifoReaderRef.FifoRef.Data[Offset_ReadStart0(_LocationVector,_RTL)];
    }

    DataType*  Loc_Base1( const vector<uint32_t>& _LocationVector ) const
    {
      return  &this->FifoReaderRef.FifoRef.Data[Offset_ReadStart1(_LocationVector)];
    }

    DataType*  Loc_End0( const vector<uint32_t>& _LocationVector, const ReadSchemaType& _RTL ) const
    {
      return  &this->FifoReaderRef.FifoRef.Data[Offset_ReadEnd0(_LocationVector,_RTL)];
    }

    DataType*  Loc_End1( const vector<uint32_t>& _LocationVector, const ReadSchemaType& _RTL ) const
    {
      return  &this->FifoReaderRef.FifoRef.Data[Offset_ReadEnd1(_LocationVector,_RTL)];
    }

    uint32_t  Loc_Num0( const vector<uint32_t>& _LocationVector, const ReadSchemaType& _RTL ) const
    {
      return  Offset_ReadEnd0( _LocationVector, _RTL ) - Offset_ReadStart0( _LocationVector, _RTL );
    }

    uint32_t  Loc_Num1( const vector<uint32_t>& _LocationVector, const ReadSchemaType& _RTL ) const
    {
      return  Offset_ReadEnd1( _LocationVector, _RTL ) - Offset_ReadStart1( _LocationVector );
    }

    /////////////////////////////
    // Offset read locations
    /////////////////////////////

    uint32_t  Offset_ReadStart0( vector<uint32_t> _LocationVector, const ReadSchemaType& _RTL ) const
    {
      _LocationVector[0] = _RTL.Start;
      return  this->FifoReaderRef.FifoRef.Extent.Offset(_LocationVector);
    }

    uint32_t  Offset_ReadEnd0( vector<uint32_t> _LocationVector, const ReadSchemaType& _RTL ) const
    {
      _LocationVector[0] = _RTL.End;
      return  this->FifoReaderRef.FifoRef.Extent.Offset(_LocationVector);
    }

    uint32_t  Offset_ReadStart1( vector<uint32_t> _LocationVector ) const
    {
      _LocationVector[0] = 0;
      return  this->FifoReaderRef.FifoRef.Extent.Offset(_LocationVector);
    }

    uint32_t  Offset_ReadEnd1( vector<uint32_t> _LocationVector, const ReadSchemaType& _RTL ) const
    {
      _LocationVector[0] = _RTL.EndWrapped;
      return  this->FifoReaderRef.FifoRef.Extent.Offset(_LocationVector);
    }

    /////////////////////////////
    // Display
    /////////////////////////////

    string  Display(void) const
    {
      ostringstream   oS;
      oS << this->FifoReaderRef.Display();
      oS << "  RWrap:" << boolalpha << this->IsFifoWrap();
      return    oS.str();
    }

    string  DisplayReadSchema( const ReadSchemaType& _ReadSchema )
    {
      ostringstream   oS;

      if(_ReadSchema.WrapFlag)
      {
        oS << "-ReadSchema: "
            << "  Start:" << setw(3)  << _ReadSchema.Start
            << "  End+1:" << setw(3)  << _ReadSchema.End
            << "  Start2:" << setw(3) << 0
            << "  End2+1:" << setw(3) << _ReadSchema.EndWrapped
            << "  ReqLen:" << setw(3)    << _ReadSchema.RequestedReadLength
            << "  ActualLen:" << setw(3)    << _ReadSchema.ActualReadLength
            << "  WrapFlag:" << boolalpha << _ReadSchema.WrapFlag;
      }
      else
      {
        oS << "-ReadSchema: "
            << "  Start:" << setw(3) << _ReadSchema.Start
            << "  End+1:" << setw(3) << _ReadSchema.End
            << "  ReqLen:" << setw(3) << _ReadSchema.RequestedReadLength
            << "  ActualLen:" << setw(3) << _ReadSchema.ActualReadLength
            << "  WrapFlag:" << boolalpha << _ReadSchema.WrapFlag;
      }

      return    oS.str();
    }
  };

public:

  typedef enum ReadTypeEnum
  {
    READ_FIXED_AMOUNT,
    READ_FIXED_AMOUNT_EXCEPT_ON_LAST,
    READ_ANY_EXCEPT_NO_DATA,
    READ_ANY
  }  ReadType;

public:

  FIFO_Reader( FIFO<DataType,AllocType>& _FIFO ) :
  FifoRef {_FIFO},
  KillFlag {false}
  {
    unique_lock<mutex>  Lock(this->Mut);

    _FIFO.RegisterListener(this);

    // Register self as a FIFO Reader
    {
      lock_guard<mutex>  Lock2(_FIFO.GetMut());

      this->NumReadable = _FIFO.GetNumReadable();
      this->Back        = _FIFO.GetBack();
      this->Proc        = _FIFO.GetProc();
    }

    Lock.unlock();
  }

  ~FIFO_Reader(void)
  { // UnRegister self as a FIFO Reader
    unique_lock<mutex>    Lock(this->Mut);

    this->FifoRef.UnRegisterListener(this);
  }

  //////////////////////////////////////////////////////////////////////////////////////
  // Accessors
  //////////////////////////////////////////////////////////////////////////////////////

  condition_variable&  GetCV(void)
  {
    return  this->CV;
  }

  mutex&  GetMut(void)
  {
    return  this->Mut;
  }

  FIFO<DataType,AllocType>&  GetFifoRef(void) const
  {
    return  this->FifoRef;
  }

  const BufMap<DataType,AllocType>&  GetBufMap(void) const
  {
    return  this->FifoRef.GetBufMap();
  }

  uint32_t  GetProc(void) const
  {
    return  this->Proc;
  }

  uint32_t  GetNumReadable(void) const
  {
    return  this->NumReadable;
  }

  //////////////////////////////////////////////////////////////////////////////////////
  // Utility
  //////////////////////////////////////////////////////////////////////////////////////

  void  SetKillFlag( const bool _KillFlag=true )
  {
    this->KillFlag = _KillFlag;

    if(this->KillFlag)
    {
      this->CV.notify_all();
    }
  }

  uint32_t  GetVectorLen(void) const
  {
    return  this->FifoRef.GetBufMap().GetDimensions().size();
  }

  inline uint32_t  Normalize( const uint32_t _Value )
  {
    return  _Value % this->FifoRef.GetBufMap().GetLengthLowestDim();
  }

  //////////////////////////////////////////////////////////////////////////////////////
  // Data Stream Flags
  //////////////////////////////////////////////////////////////////////////////////////


  bool  GetEndOfDataFlag(void) const
  {
    return  this->FifoRef.GetEndOfDataFlag();
  }

  bool  IsFifoComplete(void) const
  {
    return  this->FifoRef.IsFifoComplete();
  }

  uint32_t  GetAmountOfDataAvail(void) const
  {
    return  this->FifoRef.GetNumReadable();
  }

  //////////////////////////////////////////////////////////////////////////////////////
  // Read Requests
  //////////////////////////////////////////////////////////////////////////////////////

  void  FIFO_Read
        (
          ReadType                             ReadTypeCommand,
          const uint32_t                       RequestedReadLen,
          FifoReaderTool<DataType,AllocType>&  FRT
        )
  {
    unique_lock<mutex>  Lock(this->Mut);

    const bool  HasSourceFifoBeenClosed = this->FifoRef.GetEndOfDataFlag();

    // Get the ReadLockTool
    auto&  RLT = FRT.GetReadLockTool();

    RLT.SetRequestedReadLength(RequestedReadLen);

    if( this->KillFlag )  return;

    switch(ReadTypeCommand)
    {
      case  READ_FIXED_AMOUNT:
        this->CV.wait( Lock,
                      [&] { return  this->KillFlag || (RequestedReadLen <= this->NumReadable); } );

        RLT.SetActualReadLength(RequestedReadLen);
      break;

      case  READ_FIXED_AMOUNT_EXCEPT_ON_LAST:
        this->CV.wait( Lock, [&]{ return  this->KillFlag ||
                                          HasSourceFifoBeenClosed ||
                                          (RequestedReadLen <= this->NumReadable); } );

        if(!HasSourceFifoBeenClosed)
        {
          RLT.SetActualReadLength(RequestedReadLen);
        }
        else
        {
          RLT.SetActualReadLength(this->NumReadable);
        }
      break;

      case  READ_ANY_EXCEPT_NO_DATA:
        this->CV.wait( Lock, [&]{ return  this->KillFlag ||
                                          (HasSourceFifoBeenClosed && (this->NumReadable == 0U)) ||
                                          (this->NumReadable > 0U); } );

        RLT.SetActualReadLength(this->NumReadable);
      break;

      case  READ_ANY:
        RLT.SetActualReadLength(this->NumReadable);
      break;

      default:
        assert(false);
      break;
    }

    Lock.unlock();

    if( this->KillFlag )  return;

    // Look for exception conditions
    if(this->DebugPrint)  this->TestForException("FIFO_Read");

    this->FifoRef.NotifyOwnCV();

    // Generate and cache the write schema
    FRT.GetReadSchema_Calc();
  }

  void  FIFO_Read_End
        (
          FifoReaderTool<DataType,AllocType>&  FRT,
          uint32_t                             LengthOverride = 0
        )
  {
    assert( !this->GetEndOfDataFlag() );

    if( LengthOverride == 0 )
    { // Normal read --- Get the ReadLockTool
      const auto&  RLT = FRT.GetReadLockTool();

      this->Incr_Back_L( RLT.GetActualReadLength() );
    }
    else
    { // A read with advance (e.g. 75% overlap &c.)
      this->Incr_Back_L( LengthOverride );
    }

    // Look for exception conditions
    if(this->DebugPrint)  this->TestForException("FIFO_Read_End");

    this->FifoRef.NotifyOwnCV();
  }

  //////////////////////////////////////////////////////////////////////////////////////
  // Data copies
  //////////////////////////////////////////////////////////////////////////////////////

  // General FIFO_Reader readable pointer at location '_Loc'
  const DataType*  GetFifoAddress
   (
     const FIFO_Reader<DataType,AllocType>::ReadLockTool&                           _ReadLockTool,
     const typename FIFO_Reader<DataType,AllocType>::ReadLockTool::ReadSchemaType&  _ReadSchema,
     const vector<uint32_t>&                                                        _Loc
   )
  {
    return  _ReadLockTool.Loc_Base0( _Loc, _ReadSchema );
  }

  //////////////////////////////////////////////////////////////////////////////////////
  // Incr
  //////////////////////////////////////////////////////////////////////////////////////
  // Suffixes:
  //      L:  Lock
  //      T:  Test State
  /////////////////////////////////////////

  inline void  Incr_Proc( const uint32_t _IncrBy )
  {
    this->NumReadable  += _IncrBy;
    this->Proc          = this->Normalize( this->Proc + _IncrBy );
  }

  inline void  Incr_Proc_L( const uint32_t _IncrBy )
  {
    lock_guard<mutex>  Lock(this->Mut);
    this->Incr_Proc(_IncrBy);
  }

  void  Incr_Proc_L_T( const uint32_t _ProcIncr )
  {
    this->Incr_Proc_L(_ProcIncr);

    // Look for exception conditions
    if(this->DebugPrint)
    {
      ostringstream   oS;
      oS << "Incr_Proc_L_T(" << _ProcIncr << ")";
      this->TestForException(oS.str());
    }
  }

  inline void  Incr_Back( const uint32_t _IncrBy )
  {
    assert( this->NumReadable >= _IncrBy );

    this->NumReadable  -= _IncrBy;
    this->Back          = this->Normalize( this->Back + _IncrBy );
  }

  void  Incr_Back_L( const uint32_t _BackIncr )
  {
    lock_guard<mutex>   Lock(this->FifoRef.GetMut());

    for( auto I: this->FifoRef.GetActiveReaders() )
    {
      I->Mut.lock();
    }

    int32_t   NewDiff = 0;
    int32_t   Temp = 0;
    bool      FirstLoop = true;

    this->Incr_Back(_BackIncr);

    // Set FIFO.B to minimum Back in all FIFO_Readers
    for( auto I: this->FifoRef.GetActiveReaders() )
    {
      Temp = this->FifoRef.GetNumReadable() - I->GetNumReadable();

      if( FirstLoop )
      {
        NewDiff   = Temp;
        FirstLoop = false;
      }
      else if( Temp < NewDiff )
      {
        NewDiff = Temp;
      }
    }

    if( NewDiff > 0 )
    {
      this->FifoRef.Incr_Back(NewDiff);
    }

    for( auto I: this->FifoRef.GetActiveReaders() )
    {
      I->Mut.unlock();
    }
  }

  //////////////////////////////////////////////////////////////////////////////////////
  // Utility items
  //////////////////////////////////////////////////////////////////////////////////////

  string  Display( const bool _DisplayFIFO=false ) const
  {
    ostringstream   oS;

    oS << "FIFO_R[" << this << "]:"
       << "  P:" << this->Proc
       << "  B:" << this->Back
       << "  nReadable:" << this->NumReadable
       << "  KillFlag:" << this->KillFlag
       << endl;

    if( _DisplayFIFO )
    {
      oS << "  " << this->FifoRef.Display(false);
    }

    return    oS.str();
  }

  // Notes on write transfers into FIFOs.  It is necessary to restrict the valid
  //   write operations due to certain pathological cases.  In essence, we want to
  //   perform a write into an available section of the output FIFO.  We exclude
  //   operations that wrap around the FIFO more than once as this isn't a data
  //   transfer technique that is seen as useful.  Transfers that wrap around the
  //   end once are valid.

  template<class OutType>
  void  WriteFifoToVector
        (
          const vector<uint32_t>&   FifoStart,        // FIFO position of first sample
          const vector<uint32_t>&   FifoIncr,         // FIFO position increment
          OutType*                  OutVect,          // The destination data
          const uint32_t            TransferLength    // Length of transfer in 'DataType'
        )
  {
    assert( !this->GetEndOfDataFlag() );

    // Offsets into the FIFO
    const auto  First = this->FifoRef.GetBufMap().Offset(FifoStart);
    const auto  Incr  = this->FifoRef.GetBufMap().Offset(FifoIncr);

    // Look for the special case where the increment is one data item.
    //   In this case we do one contiguous transfer rather than multiple
    //   single-datum transfers.
    if( Incr == 1 )
    { // Data items are adjacent in memory --- faster transfers
      const uint32_t  LastPlusOne = First + TransferLength;

      if( LastPlusOne <= this->FifoRef.GetBufMap().GetNumElements() )
      {
        std::copy
        (
          &this->GetFifoRef().GetData()[First],
          &this->GetFifoRef().GetData()[LastPlusOne],
          OutVect
        );
      }
      else
      { // Two 'std::copy' operations wrapping around the end of the FIFO
        uint32_t  DistToTopOfFifo = this->GetFifoRef().GetBufMap().GetNumElements() - First;

        std::copy
        (
          &this->GetFifoRef().GetData()[First],
          &this->GetFifoRef().GetData()[this->GetFifoRef().GetBufMap().GetNumElements()],
          OutVect
        );

        std::copy
        (
          &this->GetFifoRef().GetData()[this->GetFifoRef().GetBufMap().GetNumElements()],
          &this->GetFifoRef().GetData()[LastPlusOne],
          OutVect + DistToTopOfFifo
        );
      }
    }
    else
    { // Data items are not adjacent in memory --- slower transfers
      uint32_t  Curr = First;

      for( auto K=0U ; K < TransferLength ; ++K )
      {
        OutVect[K] = this->GetFifoRef().GetData()[Curr];

        // Incr 'Curr'
        if( (K+1) < TransferLength )
        {
          Curr += Incr;

          if( Curr >= this->GetFifoRef().GetBufMap().GetNumElements() )
          {
            Curr -= this->GetFifoRef().GetBufMap().GetNumElements();
            assert( Curr < this->GetFifoRef().GetBufMap().GetNumElements() );
          }
        }
      }
    }
  }

  inline DataType  ReadValueFromFifo( const uint32_t FifoPos ) const
  {
    return  this->FifoRef.GetData()[FifoPos];
  }

  inline uint32_t  IncrFifoPos( const uint32_t FifoPos, const uint32_t FifoIncr ) const
  {
    uint32_t  Ret = FifoPos + FifoIncr;

    if( Ret > this->FifoRef.GetBufMap().GetNumElements() )
    {
      Ret -= this->FifoRef.GetBufMap().GetNumElements();
      assert( Ret < this->FifoRef.GetBufMap().GetNumElements() );
    }

    return  Ret;
  }

private:

  //////////////////////////////////////////////////////////////////////////////////////
  // Debug modes
  //////////////////////////////////////////////////////////////////////////////////////

  const bool  DebugPrint = false;

  //////////////////////////////////////////////////////////////////////////////////////
  // Access to storage FIFO
  //////////////////////////////////////////////////////////////////////////////////////

  FIFO<DataType,AllocType>&  FifoRef;

  //////////////////////////////////////////////////////////////////////////////////////
  // Synchronization and control
  //////////////////////////////////////////////////////////////////////////////////////

  atomic<bool>          KillFlag;
  mutex                 Mut;
  condition_variable    CV;

  //////////////////////////////////////////////////////////////////////////////////////
  // These Pointers are the Least Significant Digit { A, ..., Front|Back|Proc };
  //    Next Location to Write Into   ---   Front <= Proc <= Back (wrap)
  //////////////////////////////////////////////////////////////////////////////////////

  // The location of the first item available for read
  uint32_t    Back;

  // Number of items that are readable in the least significant dimension
  uint32_t    NumReadable;

  // The location of the last item available for read - 1
  uint32_t    Proc;

private:

  // Should be called with Lock on
  void  TestForException( const string& _Tag )
  {
    lock_guard<mutex>  Lock(this->FifoRef.GetMut());
    lock_guard<mutex>  Lk(this->Mut);

    if( this->FifoRef.GetProc() != this->GetProc() )
    {
      cout << endl
           << "*Error[" << __FILE__ << ":" << __LINE__ << ":" << _Tag << "]: "
           << "'this->FifoRef.Proc != this->Proc'" << endl
           << ' ' << this->Display(true);
      cout.flush();
    }

    if( this->FifoRef.GetNumReadable() < this->GetNumReadable() )
    {
      cout << endl
           << "*Error[" << __FILE__ << ":" << __LINE__ << ":" << _Tag << "]: "
           << "'this->FifoRef.NumReadable < this->NumReadable'" << endl
           << ' ' << this->Display(true);
      cout.flush();
    }
  }
};


