///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  RTIF.h                                                                            //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#pragma once

#include  <complex>
#include  <thread>
#include  <mutex>
#include  <condition_variable>
#include  <iostream>
#include  <iomanip>
#include  <atomic>
#include  "FIFO.h"
#include  "FIFO_Reader.h"

using namespace std;


extern  mutex  DebugMutex;     // Global Debug Mutex


// Rendered Template Interface
class  RTIF
{
public:

  typedef enum  eExitState
  {
    Exit_Normally,
    Exit_Due_To_Kill
  }  ExitStateType;

protected:

  enum  eState
  {
    State_Initialized,    // Execution of the constructor started
    State_Allocated,      // Execution of 'Init_Allocate' started
    State_Reset,          // Execution of 'Init_Reset' started
    State_Processing,     // Execution of 'ProcessBatches' started
    State_Kill,           // Execution of 'KillThreads' started
    State_Exit            // Execution of 'Exit' started
  }  State;

public:

  ////////////////////
  // State Accessors
  ////////////////////

  bool    IsInitialized(void) const
  {
    return  State == this->State_Initialized;
  }

  bool    IsAllocated(void) const
  {
    return  State == this->State_Allocated;
  }

  bool    IsReset(void) const
  {
    return  State == this->State_Reset;
  }

  bool    IsProcessing(void) const
  {
    return  State == this->State_Processing;
  }

  bool    IsKill(void) const
  {
    return  State == this->State_Kill;
  }

  bool    IsExit(void) const
  {
    return  State == this->State_Exit;
  }

  string    DisplayState( enum eState State )
  {
    ostringstream   oS;

    switch(State)
    {
    case  State_Initialized:
      oS << "Initialized State";
      break;

    case  State_Allocated:
      oS << "Allocated State";
      break;

    case  State_Reset:
      oS << "Reset State";
      break;

    case  State_Processing:
      oS << "Processing State";
      break;

    case  State_Kill:
      oS << "Kill State";
      break;

    case  State_Exit:
      oS << "Exit State";
      break;

    default:
      Exception   E(__FILE__,__LINE__);
      E.Msg() << "Undefined State.";
      throw  E;
      break;
    }

    return  oS.str();
  }

  // Base virtual entry points
  virtual ExitStateType  Init_Allocate(void) = 0;    // Called once to allocate dynamic memory
  virtual ExitStateType  Init_Reset(void)    = 0;    // Called any time to reset all counters to their initial state
  virtual ExitStateType  KillThreads(void)   = 0;    // Kill all threads

protected:

  // Threading and synchronization
  atomic<uint32_t>    Alive;          // A count of the number of threads that are Alive
  atomic<uint32_t>    Complete;       // A count of the number of threads that are Complete
  atomic<uint64_t>    CycleCounter;   // Rolling cycle number
  atomic<bool>        KillFlag;       // Kills threading

private:

  mutex               CvMutex;        // Mutex for the Condition Variable
  condition_variable  CV;             // The Condition Variable for basic synchronization

protected:
  // Helper routines
  void  BaseChannelNumberForThread( const unsigned int  _ThreadNumber,
                                    const unsigned int  _NumThreads,
                                    const unsigned int  _NumChannels,
                                          unsigned int& _BaseChannelForThread,
                                          unsigned int& _NumChannelsForThread );

public:

  mutex&    GetCvMutex(void)
  {
    return  this->CvMutex;
  }

  condition_variable&   GetCv(void)
  {
    return  this->CV;
  }
};


