///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  Wavefile.cpp                                                                      //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#include    <iostream>
#include    <iomanip>
#include    <chrono>
#include    <cstddef>
#include    "Wavefile.h"


WaveFileIn::WaveFileIn( const string& pFileName ) :
ValidFile(false),
BigEndianMachine(false),
InFile( ifstream( pFileName, ios::binary ) ),
NumSamples(0),
CdHeader(false),
FileName(pFileName)
{
  char  Hdr[4];

  // Make sure the file is open
  if( !InFile.good() )
  {
    Exception   E(__FILE__,__LINE__);
    E.Msg() << "Cannot open file '" << pFileName << "'.";
    throw  E;
  }

  // Set a flag indicating the endianness of this CPU
  this->BigEndianMachine = IsThisA_BigEndianMachine();

  // Mark the file position as we need to read two different types of headers:  WAV and CD
  this->FP_PreHeader = InFile.tellg();

  /////////////////////////////////////////////////////////////////////////
  // Read as WAV Header
  /////////////////////////////////////////////////////////////////////////

  // RIFF Header
  InFile.read( (char*) &Header.WavHeader.Header1, sizeof(Header.WavHeader.Header1) );

  if( InFile.gcount() != sizeof(Header.WavHeader.Header1) )
  {
    Exception   E(__FILE__,__LINE__);
    E.Msg() << "Error reading WAV File RIFF Header in WAV file.";
    throw  E;
  }

  // 'fmt ' Header
  InFile.read( (char*) &Header.WavHeader.Header2, sizeof(Header.WavHeader.Header2) );

  if( InFile.gcount() != sizeof(Header.WavHeader.Header2) )
  {
    Exception   E(__FILE__,__LINE__);
    E.Msg() << "Error reading WAV File 'fmt' Chunk in WAV file.";
    throw  E;
  }

  // Search for the 'data' segment.  In a WAV file it should start right away.  For
  //    48 KHz sampling there are other data chunks in between
  {
    // Look for 'data'
    while(true)
    {
      InFile.read( (char*) &Hdr, sizeof(Hdr) );

      if( InFile.gcount() != sizeof(Hdr) )
      {
        Exception   E(__FILE__,__LINE__);
        E.Msg() << "Error reading WAV File 'data' chunk.";
        throw  E;
      }

      if( (Hdr[0] == 'd') && (Hdr[1] == 'a') && (Hdr[2] == 't') && (Hdr[3] == 'a') )
      {
        InFile.seekg( -sizeof(Hdr), ios::cur );
        break;
      }

      // Move back 3
      InFile.seekg( -3, ios::cur );
    }

    // 'data' Chunk
    InFile.read( (char*) &Header.WavHeader.Header3, sizeof(Header.WavHeader.Header3) );

    if( InFile.gcount() != sizeof(Header.WavHeader.Header3) )
    {
      Exception   E(__FILE__,__LINE__);
      E.Msg() << "Error reading WAV File 'data; chunk in WAV file.";
      throw  E;
    }
  }

  // Mark the file position as we need to find the start of data
  this->FP_WAV_DataStart = InFile.tellg();

  // Perform any endian conversion operations necessary
  Header.WavHeader.Header1.FileLengthMinus_8 = EndianReverseOpt_u32( Header.WavHeader.Header1.FileLengthMinus_8 );
  Header.WavHeader.Header2.Sixteen = EndianReverseOpt_u16( Header.WavHeader.Header2.Sixteen );
  Header.WavHeader.Header2.PcmTag = EndianReverseOpt_u16( Header.WavHeader.Header2.PcmTag );
  Header.WavHeader.Header2.NumChannels = EndianReverseOpt_u32( Header.WavHeader.Header2.NumChannels );
  Header.WavHeader.Header2.SampleRate = EndianReverseOpt_u32( Header.WavHeader.Header2.SampleRate );
  Header.WavHeader.Header2.BytesPerSecond = EndianReverseOpt_u32( Header.WavHeader.Header2.BytesPerSecond );
  Header.WavHeader.Header2.BlockAlign = EndianReverseOpt_u16( Header.WavHeader.Header2.BlockAlign );
  Header.WavHeader.Header2.BitsPerSample = EndianReverseOpt_u16( Header.WavHeader.Header2.BitsPerSample );
  Header.WavHeader.Header3.SampleDataSize = EndianReverseOpt_u32( Header.WavHeader.Header3.SampleDataSize );

  // See if this is a WAV-type header
  this->CdHeader  = this->Header.WavHeader.Header2.Sixteen != 16;
  this->ValidFile = this->Header.WavHeader.Header2.Sixteen == 16;

  // See if we should set the 'DataStart' field
  if( this->ValidFile )
  {
    this->FP_DataStart = this->FP_WAV_DataStart;
  }

  /////////////////////////////////////////////////////////////////////////
  // If this is NOT a WAV header then try and interpret it as a CD header
  /////////////////////////////////////////////////////////////////////////

  if( !this->ValidFile )
  {
    // We must still verify this is a CD header
    this->CdHeader = false;

    // Move file back to the header start position
    InFile.seekg( this->FP_PreHeader, ios::beg );

    // Read as a CD header
    InFile.read( (char*) &Header.CdHeader.Header1, sizeof(Header.CdHeader.Header1) );

    if( InFile.gcount() != sizeof(Header.CdHeader.Header1) )
    {
      Exception   E(__FILE__,__LINE__);
      E.Msg() << "Error reading WAV File 'RIFF' Header in WAV file.";
      throw  E;
    }

    InFile.read( (char*) &Header.CdHeader.Header2, sizeof(Header.CdHeader.Header2) );

    if( InFile.gcount() != sizeof(Header.CdHeader.Header2) )
    {
      Exception   E(__FILE__,__LINE__);
      E.Msg() << "Error reading WAV File 'fmt' Header in WAV file.";
      throw  E;
    }

    // Search for the 'data' segment.  In a WAV file it should start right away.  For
    //    48 KHz sampling there are other data chunks in between
    {
      // Look for 'd' in 'data'
      while(true)
      {
        InFile.read( (char*) &Hdr, sizeof(Hdr) );

        if( InFile.gcount() != sizeof(Hdr) )
        {
          Exception   E(__FILE__,__LINE__);
          E.Msg() << "Error reading WAV File 'data' chunk.";
          throw  E;
        }

        if( (Hdr[0] == 'd') && (Hdr[1] == 'a') && (Hdr[2] == 't') && (Hdr[3] == 'a') )
        {
          InFile.seekg( -sizeof(Hdr), ios::cur );
          break;
        }

        // Move back 3
        InFile.seekg( -3, ios::cur );
      }

      // 'data' Chunk
      InFile.read( (char*) &Header.CdHeader.Header3, sizeof(Header.CdHeader.Header3) );

      if( InFile.gcount() != sizeof(Header.CdHeader.Header3) )
      {
        Exception   E(__FILE__,__LINE__);
        E.Msg() << "Error reading WAV File 'data; chunk in WAV file.";
        throw  E;
      }
    }

    // Mark the file position as we need to find the start of data
    this->FP_CD_DataStart = InFile.tellg();

    // Perform any endian conversion operations necessary
    Header.CdHeader.Header1.FileLengthMinus_8 = EndianReverseOpt_u32( Header.CdHeader.Header1.FileLengthMinus_8 );
    Header.CdHeader.Header2.Eighteen = EndianReverseOpt_u16( Header.CdHeader.Header2.Eighteen );
    Header.CdHeader.Header2.PcmTag = EndianReverseOpt_u16( Header.CdHeader.Header2.PcmTag );
    Header.CdHeader.Header2.NumChannels = EndianReverseOpt_u32( Header.CdHeader.Header2.NumChannels );
    Header.CdHeader.Header2.SampleRate = EndianReverseOpt_u32( Header.CdHeader.Header2.SampleRate );
    Header.CdHeader.Header2.BytesPerSecond = EndianReverseOpt_u32( Header.CdHeader.Header2.BytesPerSecond );
    Header.CdHeader.Header2.BlockAlign = EndianReverseOpt_u16( Header.CdHeader.Header2.BlockAlign );
    Header.CdHeader.Header2.BitsPerSample = EndianReverseOpt_u16( Header.CdHeader.Header2.BitsPerSample );
    Header.CdHeader.Header3.SampleDataSize = EndianReverseOpt_u32( Header.CdHeader.Header3.SampleDataSize );

    // See if this is a WAV-type header
    this->CdHeader  = Header.CdHeader.Header2.Eighteen == 18;
    this->ValidFile = Header.CdHeader.Header2.Eighteen == 18;

    // See if we should set the 'DataStart' field
    if( this->ValidFile )
    {
      this->FP_DataStart = this->FP_CD_DataStart;
    }
    else
    { // If we didn't see a CD header then we cannot use this header
      Exception   E(__FILE__,__LINE__);
      E.Msg() << "Bad WAV file header -- neither WAV or CD format.";
      throw  E;
    }
  }

  // Move file position appropriately
  if( this->ValidFile && this->CdHeader )
  {
    this->InFile.seekg( FP_CD_DataStart, ios::beg );
    this->NumSamples  = Header.CdHeader.Header3.SampleDataSize >> 2;
  }
  else
  {
    this->InFile.seekg( FP_WAV_DataStart, ios::beg );
    this->NumSamples  = this->Header.WavHeader.Header3.SampleDataSize >> 2;
  }
}


uint16_t    WaveFileIn::GetBitsPerSample(void) const
{
  if( this->CdHeader )
  {
    return  this->Header.CdHeader.Header2.BitsPerSample;
  }
  else
  {
    return  this->Header.WavHeader.Header2.BitsPerSample;
  }
}


uint16_t    WaveFileIn::GetBlockAlign(void) const
{
  if( this->CdHeader )
  {
    return  this->Header.CdHeader.Header2.BlockAlign;
  }
  else
  {
    return  this->Header.WavHeader.Header2.BlockAlign;
  }
}


uint32_t    WaveFileIn::GetNumSamples(void) const
{
  return  NumSamples;
}


uint16_t    WaveFileIn::GetNumChannels(void) const
{
  if( this->CdHeader )
  {
    return  this->Header.CdHeader.Header2.NumChannels;
  }
  else
  {
    return  this->Header.WavHeader.Header2.NumChannels;
  }
}


int8_t    WaveFileIn::GetSample_s8(  const uint32_t _SampleNum, const uint8_t _Channel )
{
  int8_t      R = 0;
  streampos   FP = this->FP_DataStart;

  // Offset to the data --- assume data length is a multiple of 8 bits
  FP += _SampleNum * this->GetBlockAlign();
  FP += _Channel * (this->GetBitsPerSample()>>3);

  // Position the file at the data if necessary
  if( this->InFile.tellg() != FP )
  { // Position the file at the data
    this->InFile.seekg( FP, ios::beg );
  }

  // Read the sample
  this->InFile.read( (char*) &R, sizeof(R) );

  if( this->InFile.gcount() != sizeof(int16_t) )
  {
    Exception   E(__FILE__,__LINE__);
    E.Msg() << "Unexpectedly Reached the end of file on '" << this->FileName << "'.";
    throw  E;
  }

  // As this is an 8-bit sample we must subtract 127 from it
  R -= 127;

  return    R;
}


int16_t   WaveFileIn::GetSample_s16( const uint32_t _SampleNum, const uint8_t _Channel )
{
  int16_t     R = 0;
  streampos   FP = this->FP_DataStart;

  // Offset to the data --- assume data length is a multiple of 8 bits
  FP += _SampleNum * this->GetBlockAlign();
  FP += _Channel * (this->GetBitsPerSample()>>3);

  if( this->InFile.tellg() != FP )
  { // Position the file at the data
    this->InFile.seekg( FP, ios::beg );
  }

  // Read the sample
  this->InFile.read( (char*) &R, sizeof(R) );

  if( this->InFile.gcount() != sizeof(int16_t) )
  {
    Exception   E(__FILE__,__LINE__);
    E.Msg() << "Unexpectedly Reached the end of file on '" << this->FileName << "'.";
    throw  E;
  }

  // Debug Mode
  if( this->ReadSequetialInWavSamples && (_Channel == 0) )
  {
    R = this->EndianReverseOpt_s16( _SampleNum >> 16, true );
  }
  else if( this->ReadSequetialInWavSamples && (_Channel == 1) )
  {
    R = this->EndianReverseOpt_s16( _SampleNum & 0xFFFFU, true );
  }
  else if( !this->ReadSequetialInWavSamples )
  { // Normal sample -- Correct the Endianess if necessary
    R = this->EndianReverseOpt_s16(R);
  }

  return    R;
}


// Endian conversions
uint16_t   WaveFileIn::EndianReverseOpt_u16( const uint16_t _Source, const bool _Force ) const
{
  // If this is a little endian machine no need to byte reverse
  if( !_Force && !this->BigEndianMachine )   return  _Source;

  // Big endian processing follows
  uint16_t        Out  = 0;
  uint8_t*        pOut = (uint8_t*)  &Out;
  const uint8_t*  pSource = (const uint8_t*)  &_Source;

  pOut[0] = pSource[1];
  pOut[1] = pSource[0];

  return  Out;
}


int16_t   WaveFileIn::EndianReverseOpt_s16( const int16_t _Source, const bool _Force ) const
{
  // If this is a little endian machine no need to byte reverse
  if( !_Force && !this->BigEndianMachine )   return  _Source;

  // Big endian processing follows
  int16_t          Out  = 0;
  uint8_t*         pOut = (uint8_t*)  &Out;
  const uint8_t*   pSource = (const uint8_t*)  &_Source;

  pOut[0] = pSource[1];
  pOut[1] = pSource[0];

  return  Out;
}


uint32_t   WaveFileIn::EndianReverseOpt_u32( const uint32_t _Source, const bool _Force ) const
{
  // If this is a little endian machine no need to byte reverse
  if( !_Force && !this->BigEndianMachine )   return  _Source;

  // Big endian processing follows
  uint16_t        Out  = 0;
  uint8_t*        pOut = (uint8_t*)  &Out;
  const uint8_t*  pSource = (const uint8_t*)  &_Source;

  pOut[0] = pSource[3];
  pOut[1] = pSource[2];
  pOut[2] = pSource[1];
  pOut[3] = pSource[0];

  return  Out;
}


int32_t   WaveFileIn::EndianReverseOpt_s32( const int32_t _Source, const bool _Force ) const
{
  // If this is a little endian machine no need to byte reverse
  if( !_Force && !this->BigEndianMachine )   return  _Source;

  int32_t         Out  = 0;
  uint8_t*        pOut = (uint8_t*)  &Out;
  const uint8_t*  pSource = (const uint8_t*)  &_Source;

  pOut[0] = pSource[3];
  pOut[1] = pSource[2];
  pOut[2] = pSource[1];
  pOut[3] = pSource[0];

  return  Out;
}


string  WaveFileIn::Display_Header( const uint16_t _Indent )
{
  if( this->IsValidFile() && this->IsWAV_Header() )
  {
    return  this->Display_WAV_Header(_Indent);
  }
  else if( this->IsValidFile() && this->IsCD_Header() )
  {
    return  this->Display_CD_Header(_Indent);
  }
  else
  {
    Exception   E(__FILE__,__LINE__);
    E.Msg() << "Cannot find a valid WAV File Header.";
    throw  E;
  }
}


string  WaveFileIn::Display_WAV_Header( const uint16_t _Indent )
{
  ostringstream   oS;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << "*WAV Header:  Valid: " << boolalpha << ValidFile << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -FileLengthMinus_8: " <<  Header.WavHeader.Header1.FileLengthMinus_8 << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -Sixteen: " << Header.WavHeader.Header2.Sixteen << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -PcmTag: " << Header.WavHeader.Header2.PcmTag << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -NumChannels: " << Header.WavHeader.Header2.NumChannels << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -SampleRate: " << Header.WavHeader.Header2.SampleRate << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -BytesPerSecond: " << Header.WavHeader.Header2.BytesPerSecond << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -BlockAlign: " << Header.WavHeader.Header2.BlockAlign << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -BitsPerSample: " << Header.WavHeader.Header2.BitsPerSample << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -SampleDataSize: " << Header.WavHeader.Header3.SampleDataSize << endl;

  return  oS.str();
}


string  WaveFileIn::Display_CD_Header( const uint16_t _Indent )
{
  ostringstream   oS;


  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << "*CD Header:  Valid: " << boolalpha << CdHeader << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -FileLengthMinus_8: " <<  Header.CdHeader.Header1.FileLengthMinus_8 << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -Eighteen: " << Header.CdHeader.Header2.Eighteen << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -PcmTag: " << Header.CdHeader.Header2.PcmTag << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -NumChannels: " << Header.CdHeader.Header2.NumChannels << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -SampleRate: " << Header.CdHeader.Header2.SampleRate << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -BytesPerSecond: " << Header.CdHeader.Header2.BytesPerSecond << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -BlockAlign: " << Header.CdHeader.Header2.BlockAlign << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -BitsPerSample: " << Header.CdHeader.Header2.BitsPerSample << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -SampleDataSize: " << Header.CdHeader.Header3.SampleDataSize << endl;

  return  oS.str();
}


WaveFileNew::WaveFileNew( const string&    _FileName,
                          const uint16_t   _NunChannels,
                          const uint16_t   _NumBits,
                          const uint32_t   _SampleRate    ) :
OutFile(_FileName, ios::trunc),
BigEndianMachine {this->IsThisA_BigEndianMachine()},
NumSamples {0},
NumChannels {_NunChannels},
NumBits {_NumBits},
NumBytes ( NumBits>>3 ),
SampleRate {_SampleRate},
ValidFile {OutFile.is_open()},
CdHeader {false}
{
  // Make sure the file is open
  if( !OutFile.good() )
  {
    ostringstream     oS;
    oS << "Cannot open WAV File '" << _FileName << "' as a new WAH file.";
    throw   oS.str();
  }

  // Writes place holders
  this->WriteWaveHeader();
}


void  WaveFileNew::CloseFileHeader(void)
{
  // Write the file length field
  this->OutFile.seekp( offsetof( WaveFileFormat::WaveFileHeaderType, Header1.FileLengthMinus_8 ), ios::beg );

  uint32_t  FileSizeMinus8 = sizeof(WaveFileFormat::WaveFileHeaderType) + this->NumSamples * 4 - 10;

  FileSizeMinus8 = EndianReverseOpt_u32(FileSizeMinus8);

  this->OutFile.write( (char*) &FileSizeMinus8, sizeof(FileSizeMinus8) );

  // Write the file length field
  this->OutFile.seekp( offsetof( WaveFileFormat::WaveFileHeaderType, Header3.SampleDataSize ), ios::beg );

  uint16_t    NumBytes = (this->NumBits + 7) >> 3;
  uint32_t    DataSize = this->NumChannels * NumBytes * this->NumSamples;

  DataSize = EndianReverseOpt_u32(DataSize);

  this->OutFile.write( (char*) &DataSize, sizeof(DataSize) );
}


void  WaveFileNew::WriteWaveHeader(void)
{
  uint16_t  BlockAlign  = (this->NumChannels * this->NumBits) >> 3;
  uint32_t  BytesPerSec = this->SampleRate * BlockAlign;

  this->Header.WavHeader =
                WaveFileFormat::WaveFileHeaderType
                {
                  {
                    { 'R', 'I', 'F', 'F' },
                    0,                              // Placeholder for file length field
                    { 'W', 'A', 'V', 'E' }
                  },

                  {
                    { 'f', 'm', 't', ' ' },
                    16,                               // Header length
                    1,                                // PCM tag
                    this->NumChannels,                // Number of Channels
                    this->SampleRate,                 // Sample Rate
                    BytesPerSec,                      // Sample_rate * block_align
                    BlockAlign,                       // Channels * bits/sample / 8
                    this->NumBits                     // Num Bits
                  },

                  {
                    { 'd', 'a', 't', 'a' },         // 'data'
                    0                               // Placeholder for the samples
                  }
                };

  // Perform byte reversal so we send as Little Endian
  this->Header.WavHeader.Header1.FileLengthMinus_8 = EndianReverseOpt_u32( this->Header.WavHeader.Header1.FileLengthMinus_8 );
  this->Header.WavHeader.Header2.Sixteen = EndianReverseOpt_u16( this->Header.WavHeader.Header2.Sixteen );
  this->Header.WavHeader.Header2.PcmTag = EndianReverseOpt_u16( this->Header.WavHeader.Header2.PcmTag );
  this->Header.WavHeader.Header2.NumChannels = EndianReverseOpt_u32( this->Header.WavHeader.Header2.NumChannels );
  this->Header.WavHeader.Header2.SampleRate = EndianReverseOpt_u32( this->Header.WavHeader.Header2.SampleRate );
  this->Header.WavHeader.Header2.BytesPerSecond = EndianReverseOpt_u32( this->Header.WavHeader.Header2.BytesPerSecond );
  this->Header.WavHeader.Header2.BlockAlign = EndianReverseOpt_u16( this->Header.WavHeader.Header2.BlockAlign );
  this->Header.WavHeader.Header2.BitsPerSample = EndianReverseOpt_u16( this->Header.WavHeader.Header2.BitsPerSample );
  this->Header.WavHeader.Header3.SampleDataSize = EndianReverseOpt_u32( this->Header.WavHeader.Header3.SampleDataSize );

  // Write the headers
  OutFile.write( (char*) &this->Header.WavHeader, sizeof(this->Header.WavHeader) );

  // Save the current file position as this is the start of data
  this->DataStart = OutFile.tellp();
}


void  WaveFileNew::SetSample_s16( const uint32_t  _SampleNum,
                                  const uint16_t  _ChannelNum,
                                  int16_t         _Sample      )
{
  uint32_t  Offset = (_SampleNum * this->NumChannels + _ChannelNum ) * this->NumBytes;
  streampos FP     =  this->DataStart + (streampos) Offset;

  // Debug Mode: Write sequential samples for debug purposes
  if( this->WriteSequetialInWavSamples )
  {
    switch(_ChannelNum)
    {
    case 0:
      _Sample = _SampleNum >> 16;
    break;

    case 1:
      _Sample = _SampleNum & 0xFFFFUL;
    break;

    default:
      _Sample = 0;
    break;
    }
  }

  // Handle the endianess of the sample
  _Sample = EndianReverseOpt_s16(_Sample);

  if( this->OutFile.tellp() != FP )
  {
    this->OutFile.seekp( FP, ios::beg );
  }

  this->OutFile.write( (char*) &_Sample, sizeof(_Sample) );

  // Save the largest sample
  if( _SampleNum > this->NumSamples )
  {
    this->NumSamples = _SampleNum;
  }
}


string  WaveFileNew::Display_Header( const uint16_t _Indent )
{
  if( this->IsValidFile() && this->IsWAV_Header() )
  {
    return  this->Display_WAV_Header(_Indent);
  }
  else if( this->IsValidFile() && this->IsCD_Header() )
  {
    return  this->Display_Header(_Indent);
  }
  else
  {
    Exception   E(__FILE__,__LINE__);
    E.Msg() << "Cannot find a valid WAV File Header.";
    throw  E;
  }
}


string  WaveFileNew::Display_WAV_Header( const uint16_t _Indent )
{
  ostringstream   oS;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << "*WAV Header:  Valid: " << boolalpha << ValidFile << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -FileLengthMinus_8: " <<  this->Header.WavHeader.Header1.FileLengthMinus_8 << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -Sixteen: " << this->Header.WavHeader.Header2.Sixteen << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -PcmTag: " << this->Header.WavHeader.Header2.PcmTag << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -NumChannels: " << this->Header.WavHeader.Header2.NumChannels << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -SampleRate: " << this->Header.WavHeader.Header2.SampleRate << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -BytesPerSecond: " << this->Header.WavHeader.Header2.BytesPerSecond << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -BlockAlign: " << this->Header.WavHeader.Header2.BlockAlign << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -BitsPerSample: " << this->Header.WavHeader.Header2.BitsPerSample << endl;

  for( auto K=0U ; K<_Indent ; ++K )  oS << ' ';
  oS << " -SampleDataSize: " << this->Header.WavHeader.Header3.SampleDataSize << endl;

  return  oS.str();
}


// Endian conversions
uint16_t   WaveFileNew::EndianReverseOpt_u16( const uint16_t _Source, const bool _Force ) const
{
  // If this is a little endian machine no need to byte reverse
  if( !_Force && !this->BigEndianMachine )   return  _Source;

  // Big endian processing follows
  uint16_t        Out  = 0;
  uint8_t*        pOut = (uint8_t*)  &Out;
  const uint8_t*  pSource = (const uint8_t*)  &_Source;

  pOut[0] = pSource[1];
  pOut[1] = pSource[0];

  return  Out;
}


int16_t   WaveFileNew::EndianReverseOpt_s16( const int16_t _Source, const bool _Force ) const
{
  // If this is a little endian machine no need to byte reverse
  if( !_Force && !this->BigEndianMachine )   return  _Source;

  // Big endian processing follows
  int16_t          Out  = 0;
  uint8_t*         pOut = (uint8_t*)  &Out;
  const uint8_t*   pSource = (const uint8_t*)  &_Source;

  pOut[0] = pSource[1];
  pOut[1] = pSource[0];

  return  Out;
}


uint32_t   WaveFileNew::EndianReverseOpt_u32( const uint32_t _Source, const bool _Force ) const
{
  // If this is a little endian machine no need to byte reverse
  if( !_Force && !this->BigEndianMachine )   return  _Source;

  // Big endian processing follows
  uint16_t        Out  = 0;
  uint8_t*        pOut = (uint8_t*)  &Out;
  const uint8_t*  pSource = (const uint8_t*)  &_Source;

  pOut[0] = pSource[3];
  pOut[1] = pSource[2];
  pOut[2] = pSource[1];
  pOut[3] = pSource[0];

  return  Out;
}


int32_t   WaveFileNew::EndianReverseOpt_s32( const int32_t _Source, const bool _Force ) const
{
  // If this is a little endian machine no need to byte reverse
  if( !_Force && !this->BigEndianMachine )   return  _Source;

  int32_t         Out  = 0;
  uint8_t*        pOut = (uint8_t*)  &Out;
  const uint8_t*  pSource = (const uint8_t*)  &_Source;

  pOut[0] = pSource[3];
  pOut[1] = pSource[2];
  pOut[2] = pSource[1];
  pOut[3] = pSource[0];

  return  Out;
}


