///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  Wavefile.h                                                                        //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#pragma once

#include    <string>
#include    <fstream>
#include    <cassert>
#include    <complex>
#include    <list>
#include    "Exception.h"

using namespace std;


////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//  WaveFileFormat
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

namespace    WaveFileFormat
{
  ///////////////////////////////////////////
  // Standard WAV File format
  ///////////////////////////////////////////

  typedef struct __attribute__ ((packed))
  {
    char                RIFF_Lit[4];
    uint32_t            FileLengthMinus_8;    // File length minus 8
    char                WAVE_Lit[4];
  }     WaveFileHeader1Type;

  typedef struct __attribute__ ((packed))
  {
    char              fmt__Lit[4];            // 'fmt '
    uint32_t          Sixteen;                // 16
    uint16_t          PcmTag;                 // PCM = 1
    uint16_t          NumChannels;
    uint32_t          SampleRate;             // 44100, 48000
    uint32_t          BytesPerSecond;         // sample_rate * block_align
    uint16_t          BlockAlign;             // channels * bits/sample / 8
    uint16_t          BitsPerSample;          // 8 or 16
  }     WaveFileHeader2Type;

  typedef struct __attribute__ ((packed))
  {
    char              data_Lit[4];            // 'data'
    uint32_t          SampleDataSize;         // Number of samples below which follow.
  }     WaveFileHeader3Type;

  typedef struct __attribute__ ((packed))
  {
    WaveFileHeader1Type         Header1;
    WaveFileHeader2Type         Header2;
    WaveFileHeader3Type         Header3;
  }     WaveFileHeaderType;

  ///////////////////////////////////////////
  // CD File format
  ///////////////////////////////////////////

  typedef struct __attribute__ ((packed))
  {
    char                RIFF_Lit[4];
    uint32_t            FileLengthMinus_8;    // File length minus 8
    char                WAVE_Lit[4];
  }     CdFileHeader1Type;

  typedef struct __attribute__ ((packed))
  {
    char              fmt__Lit[4];            // 'fmt '
    uint32_t          Eighteen;               // 18
    uint16_t          PcmTag;                 // PCM = 1
    uint16_t          NumChannels;
    uint32_t          SampleRate;             // 44100, 48000
    uint32_t          BytesPerSecond;         // sample_rate * block_align
    uint16_t          BlockAlign;             // channels * bits/sample / 8
    uint16_t          BitsPerSample;          // 8 or 16
    uint16_t          Zeros;                  // 2 Extra Bytes in CD Header
  }     CdFileHeader2Type;

  typedef struct __attribute__ ((packed))
  {
    char              data_Lit[4];            // 'data'
    uint32_t          SampleDataSize;         // Number of samples below which follow.
  }     CdFileHeader3Type;

  typedef struct __attribute__ ((packed))
  {
    CdFileHeader1Type         Header1;
    CdFileHeader2Type         Header2;
    CdFileHeader3Type         Header3;
  }     CdFileHeaderType;
}


////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//  class WaveFileIn
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

class       WaveFileIn
{
private:

  // Debug Mode: Read sequential samples for debug purposes
  const bool  ReadSequetialInWavSamples = false;

public:

  WaveFileIn( const string& pFileName );

  bool    IsOpen(void) const
  {
    return  this->InFile.is_open();
  }

  inline static bool   IsThisA_BigEndianMachine(void)
  {
    uint16_t    Val1 = 1;
    uint8_t*    pVal = (uint8_t*) &Val1;
    return  (pVal[0] != 1);
  }

  // Parameters
  uint16_t  GetBitsPerSample(void) const;
  uint16_t  GetBlockAlign(void) const;
  uint32_t  GetNumSamples(void) const;
  uint16_t  GetNumChannels(void) const;

  // Data recovery items
  int8_t    GetSample_s8(  const uint32_t _SampleNum, const uint8_t _Channel );
  int16_t   GetSample_s16( const uint32_t _SampleNum, const uint8_t _Channel );

  bool      IsCD_Header(void) const
  {
    return  this->CdHeader;
  }

  bool      IsValidFile(void) const
  {
    return  this->ValidFile;
  }

  bool      IsWAV_Header(void) const
  {
    return  !this->CdHeader;
  }

  // Displays
  string    Display_Header( const uint16_t _Indent=0 );
  string    Display_WAV_Header( const uint16_t _Indent=0 );
  string    Display_CD_Header( const uint16_t _Indent=0 );

private:

  bool              ValidFile;
  bool              BigEndianMachine;  // TRUE if Big Endian
  ifstream          InFile;
  uint32_t          NumSamples;
  bool              CdHeader;  // TRUE if this is a CD Header rather than a WAV header
  string            FileName;

  // File Offsets as per 'CdHeader'
  streampos  FP_PreHeader;
  streampos  FP_WAV_DataStart;    // 'CdHeader' FALSE
  streampos  FP_CD_DataStart;     // 'CdHeader' TRUE
  streampos  FP_DataStart;        // Either WAV or CD as per 'CdHeader'

  union   uHeader
  {
    WaveFileFormat::WaveFileHeaderType    WavHeader;
    WaveFileFormat::CdFileHeaderType      CdHeader;
  }   Header;

  // Endian conversions
  uint16_t    EndianReverseOpt_u16( const uint16_t _Source, const bool _Force=false ) const;
  uint32_t    EndianReverseOpt_u32( const uint32_t _Source, const bool _Force=false ) const;
  int16_t     EndianReverseOpt_s16( const int16_t  _Source, const bool _Force=false ) const;
  int32_t     EndianReverseOpt_s32( const int32_t  _Source, const bool _Force=false ) const;

protected:

  void    WriteWaveHeader(void);
};


////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//  WaveFileNew
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

class       WaveFileNew
{
private:

  // Debug Mode: Write sequential samples for debug purposes
  const bool  WriteSequetialInWavSamples = false;

public:

  WaveFileNew(  const string&    _FileName,
                const uint16_t   _NunChannels,
                const uint16_t   _NumBits,
                const uint32_t   _SampleRate );

  ~WaveFileNew(void)
  {
    this->CloseFileHeader();
  }

  bool    IsFileOpen(void) const
  {
    return  OutFile.is_open();
  };

  bool      IsCD_Header(void) const
  {
    return  this->CdHeader;
  }

  bool      IsValidFile(void) const
  {
    return  this->ValidFile;
  }

  bool      IsWAV_Header(void) const
  {
    return  !this->CdHeader;
  }

  // Displays
  string    Display_Header( const uint16_t _Indent=0 );
  string    Display_WAV_Header( const uint16_t _Indent=0 );

  // Close file for samples
  void      CloseFileHeader(void);

  // Samples
  void      SetSample_s16( const uint32_t   _SampleNum,
                           const uint16_t   _ChannelNum,
                           const int16_t    _Sample     );

private:

  ofstream        OutFile;
  bool            BigEndianMachine;  // TRUE if Big Endian
  uint32_t        NumSamples;
  uint16_t        NumChannels;
  uint16_t        NumBits;
  uint16_t        NumBytes;
  uint32_t        SampleRate;

  bool            ValidFile;
  bool            CdHeader;

  streampos       DataStart;

  union   uHeader
  {
    WaveFileFormat::WaveFileHeaderType    WavHeader;
    WaveFileFormat::CdFileHeaderType      CdHeader;
  }   Header;

  inline static bool   IsThisA_BigEndianMachine(void)
  {
    uint16_t    Val1 = 1;
    uint8_t*    pVal = (uint8_t*) &Val1;
    return  (pVal[0] != 1);
  }

  void    WriteWaveHeader(void);

  // Endian conversions
  uint16_t    EndianReverseOpt_u16( const uint16_t _Source, const bool _Force=false ) const;
  uint32_t    EndianReverseOpt_u32( const uint32_t _Source, const bool _Force=false ) const;
  int16_t     EndianReverseOpt_s16( const int16_t  _Source, const bool _Force=false ) const;
  int32_t     EndianReverseOpt_s32( const int32_t  _Source, const bool _Force=false ) const;
};


