///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  DataRef.h                                                                         //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#pragma once

#include    <complex>
#include    <string>
#include    <vector>
#include    <iostream>
#include    "DataPool.h"
#include    "DataPath.h"
#include    "DataNodeRef.h"
#include    "Exception.h"

using namespace std;


template<typename T>
class  DataRef
{
private:

  DataPath                          Path;
  DataPool::NodeRefType             NodeRef;      // Filled when Path resolved
  string                            DataString;

  static bool                       InputCharMaskInitted;
  static vector<ctype<char>::mask>  InputCharMask;

  // Native data item
  T                                 Data;

public:

  DataRef( const DataPath& _DataPath ) :
  Path(_DataPath),
  NodeRef({0,0}),
  DataString()
  {
    this->ReadDataFromTree();

    if(!InputCharMaskInitted)
    {
      PresetInputCharMask();
    }
  }

  DataRef( const T& _Value, const DataPath& _DataPath ) :
  Path(_DataPath),
  NodeRef({0,0}),
  DataString()
  {
    SetData(_Value);

    const bool  Ret  = SaveDataToTree(_Value);

    if(!Ret)
    {
      Exception   DE(__FILE__,__LINE__);
      DE.Msg() << "Cannot save data to 'DataTree' at path '" << _DataPath.GetPathString() << "'.";
      throw   DE;
    }

    if(!InputCharMaskInitted)
    {
      PresetInputCharMask();
    }
  }

  ~DataRef(void)
  {
    this->SaveDataStringToTree(DataString);
  }

  const T&    GetData(void) const
  {
    return  Data;
  }

  const DataPath&   GetPath(void) const
  {
    return  Path;
  }

  void        SetData( const T& _Data )
  {
    Data        = _Data;
    DataString  = GetDataAsString(_Data);
  }

  string    GetDataAsString(void) const
  {
    return  DataString;
  }

  void        SetDataAsString( const string& _DataString )
  {
    DataString  = _DataString;
    Data        =  GetDataFromString(_DataString);
  }

  DataNodeRef    GetDataNodeRef(void)
  {
    return   DataNodeRef( Path, NodeRef );
  }

  const DataPool::NodeRefType&   GetNode(void) const
  {
    return  NodeRef;
  }

  // Indirection is indicated when the first character is a '/'
  bool  IsPossibleIndirection(void) const
  {
    switch(DataString.length())
    {
      case  0:
        return  false;
      break;

      case  1:
        return  DataString[0] == '/';
      break;

      default:
        if( DataString[0] != '/' )  return  false;

        // Ignore '/*' as comments
        if( (DataString[0] == '/') && (DataString[1] == '*') )  return  false;
      break;
    }

    return  true;
  }

  void        PerformIndirection( const DataPath& BasePath )
  {
    const bool  GoodIndirection  = IsPossibleIndirection();

    if( !GoodIndirection )
    {
      Exception   DE(__FILE__,__LINE__);
      DE.Msg() << "Attempting indirection on a non-path '"
               << DataString << "'.";
      throw   DE;
    }

    // Perform the indirection
    Path  = BasePath + DataPath(DataString);

    // Must clear the Node structure to force a re-lookup
    NodeRef.Node   = nullptr;
    NodeRef.Tree   = nullptr;

    bool Exists  = ReadDataFromTree();

    if( !Exists )
    {
      Exception   DE(__FILE__,__LINE__);
      DE.Msg() << "Cannot find reference '"
               << Path.GetPathString() << "'.";
      throw   DE;
    }
  }

  void        ForceTreeFileWrite(void)
  {
    DataPool::WriteTreeToDisk(NodeRef.Tree);
  }

private:

  void        PresetInputCharMask(void)
  {
    InputCharMask[' ']   ^= ctype_base::space;
    InputCharMask['\t']  ^= ctype_base::space;
    InputCharMask['\n']  ^= ctype_base::space;
    InputCharMask['\r']  ^= ctype_base::space;
    InputCharMaskInitted  = true;
  }

  bool    ReadDataFromTree(void)
  {
    this->ReadDataStringFromTree();

    this->Data  = GetDataFromString( this->DataString );

    return  true;
  }

  void    ReadDataStringFromTree(void)
  {
    if( !this->NodeRef.IsValid() )
    { // Find by path
      this->NodeRef = wej_data_pool::DataTrees.FindPath( this->Path );
    }

    // Make sure we found the string in the tree
    if( !this->NodeRef.IsValid() )
    {
      Exception   DE(__FILE__,__LINE__);
      DE.Msg() << "Cannot find XML reference '" << Path.GetPathString() << "'.";
      throw   DE;
    }

    // Look up using the NodeRef
    this->DataString = wej_data_pool::DataTrees.Read(this->NodeRef);
  }

  void   SaveDataStringToTree( const string& _DataString )
  {
    if( !this->NodeRef.IsValid() )
    { // Find by path
      this->NodeRef = wej_data_pool::DataTrees.FindCreatePath(this->Path);
    }

    // Make sure we found the string in the tree
    if( !this->NodeRef.IsValid() )
    {
      Exception   DE(__FILE__,__LINE__);
      DE.Msg() << "Cannot find reference '" << Path.GetPathString() << "'.";
      throw   DE;
    }

    // Look up using the NodeRef
    wej_data_pool::DataTrees.Write( this->NodeRef, _DataString );
  }

  bool   SaveDataToTree( const T& _Data )
  {
    string    LocDataString = GetDataAsString(_Data);
    return    SaveDataStringToTree( LocDataString );
  }

  static string   GetDataAsString( const T& _Data );
  static T        GetDataFromString( const string& _DataString );
};


///////////////////////////////////////////////////////////////////////////////////////////////////
// NOTE:  Any definitions that use a template value <T> must be defined in the .h file.
// ====   For <> type definitions they are placed in the conjoined .cpp file.
///////////////////////////////////////////////////////////////////////////////////////////////////

// Used to for formatted input to ignore spaces and other whitespace in inputs
template<typename T>
bool      DataRef<T>::InputCharMaskInitted = false;

template<typename T>
vector<ctype<char>::mask>   DataRef<T>::InputCharMask( ctype<char>::classic_table(),
                                                       ctype<char>::classic_table() + ctype<char>::table_size );


template<typename T>
string    DataRef<T>::GetDataAsString( const T& _Data )
{
  ostringstream   oS;
  oS << setprecision(32) << _Data;
  return  oS.str();
}


template<typename T>
T     DataRef<T>::GetDataFromString( const string& _DataString )
{
  istringstream     iS(_DataString);
  T                 Temp;

  // Modify so we skip spaces and whitespace
  iS.imbue( locale(iS.getloc(), new ctype<char>(InputCharMask.data())) );
  iS.unsetf(ios::skipws);

  // Grab the data
  iS >> Temp;

  return  Temp;
}


