///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  KleinWrapperObj.cpp                                                               //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#include    <iostream>
#include    <iomanip>
#include    <typeinfo>
#include    "KleinWrapperObj.h"
#include    "CharacterObj.h"
#include    "LiteralObj.h"
#include    "OutputListObj.h"
#include    "AndRuleObj.h"

using namespace std;


namespace   wej_parser
{
///////////////////////////////////////////////////////////////////////////////////////////////////
// string   KleinWrapperObj::Display
///////////////////////////////////////////////////////////////////////////////////////////////////

string    KleinWrapperObj::Display( bool , const unsigned int  ) const
{
  ostringstream        oS;

  // Display the base item
  oS << Child->Display();

  // Display the modifiers
  oS << DisplayFlags();

  // Display the Klein eclictic
  switch(Klein)
  {
    case    NoKlien:
    break;

    case    KlienStar:
      oS << "*";
    break;

    case    KlienPlus:
      oS << "+";
    break;

    case    KlienQuestion:
      oS << "?";
    break;

    default:
      assert(false);
    break;
  }

  return  oS.str();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// bool   KleinWrapperObj::Parse
///////////////////////////////////////////////////////////////////////////////////////////////////

bool  KleinWrapperObj::Parse( ParseFrame PF )
{
  unsigned int    Count;
  bool            Ret;

  // Set rule
  PF.SetRule(this);

  for( Count=0, Ret=true ; !PF.IsInEof() ; ++Count )
  {
    // Parse character
    Ret = Child->Parse(PF);

    if(Ret)
    {
      PF.AcceptAll();
    }
    else
    {
      break;
    }

    // Single try?
    if( Klein == NoKlien )          break;
    if( Klein == KlienQuestion )    break;
  }

  switch(Klein)
  {
    case    NoKlien:
    break;

    case    KlienQuestion:
    case    KlienStar:
      Ret = true;
    break;

    case    KlienPlus:
      Ret = Count > 0;
    break;

    default:
      assert(false);
    break;
  }

  if(GrammarAbs::DebugParse)
  {
    cout << ">>KLEIN        "
         << "  [" << PF.GetRuleDebugDescriptor() << "]"
         << "  [" << PF.GetInPos() << "," << PF.GetOutListPos() << "]"
         << endl;
  }

  return    Ret;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// string   KleinWrapperObj::GetOutput
///////////////////////////////////////////////////////////////////////////////////////////////////

string  KleinWrapperObj::GetOutput( ParseFrame& PF, const unsigned int RuleNum )
{
  assert( PF.PosIn >= PF.PosIn0 );
  assert(Child);

  return    this->Child->GetOutput( PF, RuleNum );
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// void   KleinWrapperObj::GenerateCppParseTree
///////////////////////////////////////////////////////////////////////////////////////////////////

void  KleinWrapperObj::GenerateCppParseTree( IBnfToCppCov& OutputData, bool KleinLayer, string& Includes )
{
  ostringstream     oS;

  switch(Klein)
  {
    case    GrammarAbs::NoKlien:
      KleinLayer  = false;
    break;

    case    GrammarAbs::KlienStar:
      KleinLayer  = true;
      oS << "  KLEIN    = new KleinWrapperObj;" << endl;
      oS << "  KLEIN->SetKlein(GrammarAbs::KlienStar);" << endl;
      oS << "  AND->PushBack(KLEIN);" << endl;
    break;

    case    GrammarAbs::KlienPlus:
      KleinLayer  = true;
      oS << "  KLEIN    = new KleinWrapperObj;" << endl;
      oS << "  KLEIN->SetKlein(GrammarAbs::KlienPlus);" << endl;
      oS << "  AND->PushBack(KLEIN);" << endl;
    break;

    case    GrammarAbs::KlienQuestion:
      KleinLayer  = true;
      oS << "  KLEIN    = new KleinWrapperObj;" << endl;
      oS << "  KLEIN->SetKlein(GrammarAbs::KlienQuestion);" << endl;
      oS << "  AND->PushBack(KLEIN);" << endl;
    break;

    default:
      KleinLayer  = false;
      assert(false);
    break;
  }

  // Accumulate Outputs
  OutputData.Subroutines.back().Contents  += oS.str();

  // Process the child first
  Child->GenerateCppParseTree( OutputData, KleinLayer, Includes );
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// bool   KleinWrapperObj::InverseParse
///////////////////////////////////////////////////////////////////////////////////////////////////

bool  KleinWrapperObj::InverseParse( InverseParseFrame IPF )
{
  unsigned int  Count;
  bool          R;

  // Set rule
  IPF.SetRule(this);

  if( IPF.IsDebug() )
  {
    cout << ">> IParse KLEIN[" << this->Klein << ": " << IPF.Display() << "]" << endl;
  }

  for( Count=0, R=true ; !IPF.IsInEof() ; ++Count )
  {
    // Parse character
    R  = Child->InverseParse(IPF);

    if(R)
    {
      IPF.AcceptAll();
    }
    else
    {
      break;
    }

    // Single try?
    if( Klein == NoKlien )          break;
    if( Klein == KlienQuestion )    break;
  }

  switch(Klein)
  {
    case    NoKlien:
    break;

    case    KlienQuestion:
    case    KlienStar:
      R  = true;
    break;

    case    KlienPlus:
      R = Count > 0;
    break;

    default:
      assert(false);
    break;
  }

  if( IPF.IsDebug() )
  {
    const char  Ret = R ? '+' : '-';

    cout << "<<" << Ret << "IParse KLEIN[" << IPF.Display() << "]" << endl;
  }

  return    R;
}
}   // End namespace 'wej_parser'


