///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  BufMap.h                                                                          //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#pragma once

#include  <cassert>
#include  <cstdint>
#include  <sstream>
#include  <string>
#include  <vector>
#include  <list>
#include  <atomic>

using namespace std;


template<typename DataType,typename AllocType>
class   BufMap
{
public:

  static const uint32_t   SizeRatio = sizeof(DataType)/sizeof(AllocType);

public:

  BufMap( const vector<uint32_t>&  _Dimensions,
          const vector<bool>&      _WrapFlags  ) :
  Dimensions(_Dimensions),
  WrapFlags(_WrapFlags),
  NumElements(this->CalcNumElements()),
  LengthLowestDim(this->CalcLengthLowestDim()),
  NumChannels(this->NumElements/this->LengthLowestDim)
  {
    assert(_Dimensions.size() > 0);
    assert(_WrapFlags.size()  > 0);
    assert(_Dimensions.size() == _WrapFlags.size());
    assert(sizeof(DataType) >= sizeof(AllocType));
  }

  ///////////////////
  // Accessors
  ///////////////////

  uint32_t  GetLengthLowestDim(void) const
  {
    return  this->LengthLowestDim;
  }

  uint32_t  GetNumElements(void) const
  {
    return  this->NumElements;
  }

  const vector<uint32_t>&  GetDimensions(void) const
  {
    return  this->Dimensions;
  }

  ///////////////////
  // Utility
  ///////////////////

  uint32_t  Offset( const vector<uint32_t>& Pos ) const
  {
    assert(Pos.size() == this->Dimensions.size());

    // For: this->Dimensions.size()-1
    uint32_t   Ret  = Pos[this->Dimensions.size()-1];
    uint32_t   Mult = this->Dimensions[this->Dimensions.size()-1];

    // For: this->Dimensions.size()-2 to 1
    for( uint32_t K = this->Dimensions.size()-2 ; K > 0 ; --K )
    {
      Ret  += Pos[K] * Mult;
      Mult *= this->Dimensions[K];
    }

    Ret += Pos[0] * Mult;

    return  Ret * BufMap::SizeRatio;
  }

  void  SetLowestWrapFlag( const bool _WrapFlag )
  {
    this->WrapFlags[0] = _WrapFlag;
  }

  uint32_t  GetNumDimensions(void) const
  {
    return  Dimensions.size();
  }

  string  DisplayPos( const vector<uint32_t>& _Pos ) const
  {
    ostringstream   oS;

    oS << this->DisplayIntVector(_Pos) << "  ";
    oS << this->DisplayIntVector(Dimensions) << "  ";
    oS << "  offset:" << this->Offset(_Pos);

    return    oS.str();
  }

  string  DisplayIntVector( const vector<uint32_t>&  _Vect ) const
  {
    ostringstream   oS;

    oS << "[";

    for( auto K=0U ; K < _Vect.size() ; ++K )
    {
      oS << _Vect[K];
      if( K != (_Vect.size()-1) )   oS << ',';
    }

    oS << "]";

    return  oS.str();
  }

  string  DisplayBoolVector( const vector<bool>&  _Vect ) const
  {
    ostringstream   oS;

    oS << "[";

    for( auto K=0U ; K < _Vect.size() ; ++K )
    {
      if( _Vect[K] )
      {
        oS << "T";
      }
      else
      {
        oS << "F";
      }

      if( K != (_Vect.size()-1) )   oS << ',';
    }

    oS << "]";

    return  oS.str();
  }

  string  Display(void) const
  {
    ostringstream   oS;

    oS << "*BufMap:  ";
    oS << "Dim:" << this->DisplayIntVector(this->Dimensions) << "  ";
    oS << "Wrap:" << this->DisplayBoolVector(this->WrapFlags) << "  ";
    oS << "NumElements:" << this->NumElements << "  ";
    oS << "LengthLowestDim:" << this->LengthLowestDim << "  ";
    oS << "NumChannels:" << this->NumChannels;

    return    oS.str();
  }

private:

  uint32_t  CalcNumElements(void) const
  {
    uint32_t  Prod = 1;

    for( auto K=0U ; K < Dimensions.size() ; ++K )
    {
      Prod *= this->Dimensions[K];
    }

    return    Prod;
  }

  uint32_t  CalcLengthLowestDim(void) const
  {
    return    this->Dimensions[0];
  }

private:

  // Base values
  vector<uint32_t>  Dimensions;
  vector<bool>      WrapFlags;

  // Calculated Values
  uint32_t          NumElements;
  uint32_t          LengthLowestDim;
  uint32_t          NumChannels;
};


