///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  EFAPI.h                                                                           //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#pragma once

#include  <complex>
#include  <cstdint>
#include  <cmath>
#include  <vector>
#include  "AlignedMemAlloc.h"

using namespace std;

//////////////////////////////////////////////////////////////////////////////////////
// Complex-valued FAPI
/////////////////////////////////////////////////////////////////////////////////////
// EFAPI_Complex<InType,ProcType,OutType,N_Len,R_Len>  EFAPI_Complex( BetaParam )
//////////////////////////////////////////////////////////////////////////////////////

template
<
  typename  InType,
  typename  ProcType,
  typename  OutType,
  uint32_t  N_Len,
  uint32_t  R_Len
>
class  EFAPI_Complex
{
public:

  // Items used for transfer data from this class
  typedef AVX_Vector<AVX_Vector<ProcType>>  MatrixType;
  typedef AVX_Vector<ProcType>              VectorType;

  const uint32_t  RbyRmatrixSize = R_Len * R_Len;
  const uint32_t  WbyRmatrixSize = N_Len * R_Len;

public:

  EFAPI_Complex(void) :
  Beta(0),
  X( N_Len, 0 ),
  W( N_Len, AVX_Vector<ProcType>(R_Len, 0) ),
  R( R_Len, AVX_Vector<ProcType>(R_Len, 0) ),
  y( R_Len, 0 ),
  h( R_Len, 0 ),
  g( R_Len, 0 ),
  Z( R_Len, AVX_Vector<ProcType>(R_Len,0) ),
  EpsSqr( 0 ),
  Tau( 0 ),
  Eta( 0 ),
  yp( R_Len, 0 ),
  hp( R_Len, 0 ),
  Eps( R_Len, 0 ),
  ep( N_Len, 0 ),
  Vect_N_1( N_Len, 0 ),
  Vect_R_1( R_Len, 0 ),
  Vect_R_2( R_Len, 0 ),
  Vect_R_3( N_Len, 0 ),
  Mat_N_R_1( N_Len, AVX_Vector<ProcType>(R_Len, 0) ),
  Var_1(0)
  {
  }

  EFAPI_Complex( const ProcType& BetaParam ) :
  Beta(BetaParam),
  X( N_Len, 0 ),
  W( N_Len, AVX_Vector<ProcType>(R_Len, 0) ),
  R( R_Len, AVX_Vector<ProcType>(R_Len, 0) ),
  y( R_Len, 0 ),
  h( R_Len, 0 ),
  g( R_Len, 0 ),
  Z( R_Len, AVX_Vector<ProcType>(R_Len,0) ),
  EpsSqr( 0 ),
  Tau( 0 ),
  Eta( 0 ),
  yp( R_Len, 0 ),
  hp( R_Len, 0 ),
  Eps( R_Len, 0 ),
  ep( N_Len, 0 ),
  Vect_N_1( N_Len, 0 ),
  Vect_R_1( R_Len, 0 ),
  Vect_R_2( R_Len, 0 ),
  Vect_R_3( N_Len, 0 ),
  Mat_N_R_1( N_Len, AVX_Vector<ProcType>(R_Len, 0) ),
  Var_1(0)
  {
    this->Init(BetaParam);
  }

  void  Init( const ProcType& BetaParam )
  {
    this->Beta = BetaParam;

    // Init W
    for( auto K=0U ; K < N_Len ; ++K )
    {
      for( auto L=0U ; L < R_Len ; ++L )
      {
        this->W[K][L] = (K == L) ? 1 : 0;
      }
    }

    // Init Z
    for( auto K=0U ; K < R_Len ; ++K )
    {
      for( auto L=0U ; L < R_Len ; ++L )
      {
        this->Z[K][L] = (K == L) ? 1 : 0;
      }
    }
  }

  uint32_t  GetDataLen(void) const
  {
    return  N_Len;
  }

  uint32_t  GetNumEigenvectors(void) const
  {
    return  R_Len;
  }

  /////////////////
  // Accessors
  /////////////////

  // Vector size is 'R_Len'
  const VectorType&  GetEigenVector( const uint32_t NumParam ) const
  {
    return  this->W[NumParam];
  }

  // Matrix size is 'N_Len * R_Len'
  const MatrixType&  GetEigenMatrix(void) const
  {
    return  this->W;
  }

  uint32_t  GetEigenMatrixSize(void) const
  {
    return  N_Len * R_Len;
  }

  const ProcType&  GetEigenValue( const uint32_t NumParam ) const
  {
    return  this->Z[NumParam][NumParam];
  }

  // Matrix size is 'R_Len * R_Len'
  const MatrixType&  GetEigenValueMatrix(void) const
  {
    return  this->Z;
  }

  uint32_t  GetEigenValueMatrixSize(void) const
  {
    return  R_Len * R_Len;
  }

  /////////////////
  // Processing
  /////////////////

  void  DataStep( const AVX_Vector<InType>& In )
  {
    assert( In.size() == N_Len );

    // Convert data to processing type
    for( auto K=0U ; K < N_Len ; ++K )
    {
      this->X[K] = In[K];
    }

    // [ 1]  y = W[h] * x
    this->MatVectMult_MhV_N( this->W, this->X, this->y );

    // [ 2]  h = Z * y
    this->MatVectMult_MV_N( this->Z, this->y, this->h );

    // [ 3]  g = h / ( beta + y[H] * h )
    assert( (this->Beta + DotProduct_VhV_N( this->y, this->h )) != 0. );

    this->VectScalarMult_VN_V
    (
      this->h,
      1. / ( this->Beta + DotProduct_VhV_N( this->y, this->h ) ),
      this->g
    );

    // [ 4]  EpsSqr = NormSqr(x) - NormSqr(y)
    this->EpsSqr = this->VectMag2_VhV_N(this->X) - this->VectMag2_VhV_N(this->y);

    // [ 5]  Tau = epsSqr / (1 + EpsSqr * Norm2(g) + sqrt( 1 + EpsSqr * Norm2(g)))
    const ProcType  Norm2G = this->VectMag2_VhV_N(this->g);
    this->Tau = this->EpsSqr / ( 1. + this->EpsSqr * Norm2G + sqrt( 1. + this->EpsSqr * Norm2G ));

    // [ 6]  Eta = 1 - Tau * Norm2(g)
    this->Eta = 1. - this->Tau * Norm2G;

    // [ 7]  Yp = Eta * y + Tau * g
    this->VectScalarMult_VN_V( this->y, this->Eta, this->Vect_R_1 );
    this->VectScalarMult_VN_V( this->g, this->Tau, this->Vect_R_2 );
    this->VectAdd_VV_V( this->Vect_R_1, this->Vect_R_2, this->yp );

    // [ 8]  hp = Z[H] * yp
    this->MatVectMult_MhV_N( this->Z, this->yp, this->hp );

    // [ 9]  Eps
    this->MatVectMult_MV_V( this->Z, this->g, this->Vect_R_1 );
    this->Var_1 = DotProduct_VhV_N( this->hp, this->g );
    this->VectScalarMult_VN_V( this->g, this->Var_1, this->Vect_R_2 );
    this->VectSub_VV_V( this->Vect_R_1, this->Vect_R_2, this->Eps );
    this->VectScalarMult_VN_X( this->Eps, (this->Tau/this->Eta) );

    // [10]  Z = (1/Beta) * ( Z - g * hp[H] + Eps * g[H] )
    this->VectOuterProdAndMatSub_VVh_Ms( this->g, this->hp, this->Z );
    this->VectOuterProdAndMatAdd_VVh_Ms( this->Eps, this->g, this->Z );
    this->MatScalarMult_MN_X( this->Z, (1./this->Beta) );

    // [11]  ep = Eta * x - W * yp
    this->VectScalarMult_VN_V( this->X, this->Eta, this->Vect_N_1 );
    this->MatVectMult_MV_V( this->W, this->yp, this->Vect_R_3 );
    this->VectSub_VV_V( this->Vect_N_1, this->Vect_R_3, this->ep );

    // [12]  W = W + ep * g[H]
    this->VectOuterProdAndMatAdd_VVh_Ms( this->ep, this->g, this->W );
  }

  //////////////////////
  // Signal Processing
  //////////////////////

  // Not Thread Safe with rest of its class
  // -Project into the spanned sub-space
  void  ProjectVectIntoSubSpace( const AVX_Vector<InType>&  SourceParam,
                                       AVX_Vector<InType>&  OutputParam  )
  {
    assert( SourceParam.size() == N_Len );
    assert( OutputParam.size() == N_Len );

    // Temp_1:  W[H] * SourceParam
    for( auto K=0U ; K < R_Len ; ++K )
    {
      this->Vect_R_1[K] = 0;

      for( auto L=0U ; L < N_Len ; ++L )
      {
        this->Vect_R_1[K] += conj(this->W[L][K]) * SourceParam[L];
      }
    }

    // Temp_2:  Temp_1 * W
    for( auto K=0U ; K < N_Len ; ++K )
    {
      OutputParam[K] = 0;

      for( auto L=0U ; L < R_Len ; ++L )
      {
        OutputParam[K] += this->W[K][L] * this->Vect_R_1[L];
      }
    }
  }

  // Project into the PERP spanned sub-space
  // -ProjectVectIntoSubSpace + ProjectVectIntoPerpSpace = Identity
  void  ProjectVectIntoPerpSpace( const AVX_Vector<InType>&  SourceParam,
                                        AVX_Vector<InType>&  OutputParam  )
  {
    assert( SourceParam.size() == N_Len );
    assert( OutputParam.size() == N_Len );

    this->ProjectVectIntoSubSpace( SourceParam, OutputParam  );

    // Calculate:  I - P
    for( auto K=0U ; K < N_Len ; ++K )
    {
      OutputParam[K]  -= SourceParam[K];
    }
  }

private:

  // Parameters
  ProcType                          Beta;       // S

  // Variables
  AVX_Vector<ProcType>              X;          // N          (N == N_Len)
  AVX_Vector<AVX_Vector<ProcType>>  W;          // N * R      (R == R_Len)
  AVX_Vector<AVX_Vector<ProcType>>  R;          // R * R
  AVX_Vector<ProcType>              y;          // R
  AVX_Vector<ProcType>              h;          // R
  AVX_Vector<ProcType>              g;          // R
  AVX_Vector<AVX_Vector<ProcType>>  Z;          // R * R
  ProcType                          EpsSqr;     // S
  ProcType                          Tau;        // S
  ProcType                          Eta;        // S
  AVX_Vector<ProcType>              yp;         // R
  AVX_Vector<ProcType>              hp;         // R
  AVX_Vector<ProcType>              Eps;        // R
  AVX_Vector<ProcType>              ep;         // N

  // New working variables
  AVX_Vector<ProcType>              Vect_N_1;   // [*]  N
  AVX_Vector<ProcType>              Vect_R_1;   // [*]  R
  AVX_Vector<ProcType>              Vect_R_2;   // [*]  R
  AVX_Vector<ProcType>              Vect_R_3;   // [*]  N
  AVX_Vector<AVX_Vector<ProcType>>  Mat_N_R_1;  // [*]  N * R
  ProcType                          Var_1;      // [*]  S

private:

  ProcType  DotProduct_VhV_N( const AVX_Vector<ProcType>& U,
                              const AVX_Vector<ProcType>& V   ) const
  {
    assert( U.size() == V.size() );

    ProcType  Ret = 0;

    for( auto K=0U ; K < U.size() ; ++K )
    {
      Ret  += conj(U[K]) * V[K];
    }

    return  Ret;
  }

  ProcType  VectMag2_VhV_N( AVX_Vector<ProcType>& V ) const
  {
    ProcType  Ret = 0;

    for( auto K=0U ; K < V.size() ; ++K )
    {
      Ret  += norm(V[K]);
    }

    return  Ret;
  }

  void  MatVectMult_MhV_N(
                         const AVX_Vector<AVX_Vector<ProcType>>&  Mat,
                         const AVX_Vector<ProcType>&          Vect,
                         AVX_Vector<ProcType>&          VectOut
                       ) const
  {
    assert( Mat.size() == Vect.size() );

    for( auto K=0U ; K < Mat[0].size() ; ++K )
    {
      VectOut[K] = 0;

      for( auto L=0U ; L < Mat.size() ; ++L )
      {
        VectOut[K] += conj(Mat[L][K]) * Vect[L];
      }
    }
  }

  void  MatVectMult_MV_N(
                         const AVX_Vector<AVX_Vector<ProcType>>&  Mat,
                         const AVX_Vector<ProcType>&          Vect,
                         AVX_Vector<ProcType>&          VectOut
                       ) const
  {
    assert( Mat[0].size() == Vect.size() );
    assert( Mat.size() == VectOut.size() );

    for( auto K=0U ; K < Mat.size() ; ++K )
    {
      VectOut[K] = 0;

      for( auto L=0U ; L < Mat[K].size() ; ++L )
      {
        VectOut[K] += Mat[K][L] * Vect[L];
      }
    }
  }

  void  MatMatAdd_MM_M(
                         const AVX_Vector<AVX_Vector<ProcType>>&  Mat1,
                         const AVX_Vector<AVX_Vector<ProcType>>&  Mat2,
                         AVX_Vector<AVX_Vector<ProcType>>&  MatOut
                       ) const
  {
    assert( Mat1.size() == Mat2.size() );
    assert( Mat1[0].size() == Mat2[0].size() );
    assert( Mat1.size() == MatOut.size() );
    assert( Mat1[0].size() == MatOut[0].size() );

    for( auto K=0U ; K < Mat1.size() ; ++K )
    for( auto L=0U ; L < Mat1[0].size() ; ++L )
    {
      MatOut[K][L] = Mat1[K][L] + Mat2[K][L];
    }
  }

  void  MatMatSub_MM_M(
                         const AVX_Vector<AVX_Vector<ProcType>>&  Mat1,
                         const AVX_Vector<AVX_Vector<ProcType>>&  Mat2,
                         AVX_Vector<AVX_Vector<ProcType>>&  MatOut
                       ) const
  {
    assert( Mat1.size() == Mat2.size() );
    assert( Mat1[0].size() == Mat2[0].size() );
    assert( Mat1.size() == MatOut.size() );
    assert( Mat1[0].size() == MatOut[0].size() );

    for( auto K=0U ; K < Mat1.size() ; ++K )
    for( auto L=0U ; L < Mat1[0].size() ; ++L )
    {
      MatOut[K][L] = Mat1[K][L] - Mat2[K][L];
    }
  }

  void  MatScalarMult_MN_X(
      AVX_Vector<AVX_Vector<ProcType>>&  Mat,
                         const ProcType&            Scalar
                       ) const
  {
    for( auto K=0U ; K < Mat.size() ; ++K )
    for( auto L=0U ; L < Mat[L].size() ; ++L )
    {
      Mat[K][L] *= Scalar;
    }
  }

  void  VectScalarMult_VN_V(
                         const AVX_Vector<ProcType>&  Vect,
                         const ProcType&          Scalar,
                         AVX_Vector<ProcType>&  VectOut
                       ) const
  {
    assert( Vect.size() == VectOut.size() );

    for( auto K=0U ; K < Vect.size() ; ++K )
    {
      VectOut[K] = Scalar * Vect[K];
    }
  }

  void  VectScalarMult_VN_X(
      AVX_Vector<ProcType>&  Vect,
                         const ProcType&          Scalar
                       ) const
  {
    for( auto K=0U ; K < Vect.size() ; ++K )
    {
      Vect[K] *= Scalar;
    }
  }

  void  VectSub_VV_V(
                      const AVX_Vector<ProcType>&  Vect1,
                      const AVX_Vector<ProcType>&  Vect2,
                      AVX_Vector<ProcType>&  VectOut
                    ) const
  {
    assert( Vect1.size() == Vect2.size() );
    assert( Vect1.size() == VectOut.size() );

    for( auto K=0U ; K < Vect1.size() ; ++K )
    {
      VectOut[K] = Vect1[K] - Vect2[K];
    }
  }

  void  VectAdd_VV_V(
                      const AVX_Vector<ProcType>&  Vect1,
                      const AVX_Vector<ProcType>&  Vect2,
                      AVX_Vector<ProcType>&  VectOut
                    ) const
  {
    assert( Vect1.size() == Vect2.size() );
    assert( Vect1.size() == VectOut.size() );

    for( auto K=0U ; K < Vect1.size() ; ++K )
    {
      VectOut[K]  = Vect1[K] + Vect2[K];
    }
  }

  void  VectOuterProd_VVh_M(
                      const AVX_Vector<ProcType>&          Vect1,
                      const AVX_Vector<ProcType>&          Vect2,
                      AVX_Vector<AVX_Vector<ProcType>>&  Out
                    ) const
  {
    assert( Vect1.size() == Out.size() );
    assert( Vect2.size() == Out[0].size() );

    for( auto K=0U ; K < Vect1.size() ; ++K )
    for( auto L=0U ; L < Vect2.size() ; ++L )
    {
      Out[K][L]  = Vect1[K] * conj(Vect2[L]);
    }
  }

  void  VectOuterProdAndMatSub_VVh_Ms(
                      const AVX_Vector<ProcType>&         Vect1,
                      const AVX_Vector<ProcType>&         Vect2,
                      AVX_Vector<AVX_Vector<ProcType>>&   Out
                    )
  {
    assert( Vect1.size() == Out.size() );
    assert( Vect2.size() == Out[0].size() );

    for( auto K=0U ; K < Vect1.size() ; ++K )
    for( auto L=0U ; L < Vect2.size() ; ++L )
    {
      Out[K][L]  -= Vect1[K] * conj(Vect2[L]);
    }
  }

  void  VectOuterProdAndMatAdd_VVh_Ms(
                      const AVX_Vector<ProcType>&         Vect1,
                      const AVX_Vector<ProcType>&         Vect2,
                      AVX_Vector<AVX_Vector<ProcType>>&   Out
                    )
  {
    assert( Vect1.size() == Out.size() );
    assert( Vect2.size() == Out[0].size() );

    for( auto K=0U ; K < Vect1.size() ; ++K )
    for( auto L=0U ; L < Vect2.size() ; ++L )
    {
      Out[K][L]  += Vect1[K] * conj(Vect2[L]);
    }
  }
};


//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////
//  Real-Valued FAPI
/////////////////////////////////////////////////////////////////////////////////////
//  EFAPI_Real<InType,ProcType,OutType,N_Len,R_Len>  EFAPI_Real( BetaParam )
//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

template
<
  typename  InType,
  typename  ProcType,
  typename  OutType,
  uint32_t  N_Len,
  uint32_t  R_Len
>
class  EFAPI_Real
{
public:

  // Items used for transfer data from this class
  typedef AVX_Vector<AVX_Vector<ProcType>>  MatrixType;
  typedef AVX_Vector<ProcType>              VectorType;

  const uint32_t  RbyRmatrixSize = R_Len * R_Len;
  const uint32_t  WbyRmatrixSize = N_Len * R_Len;

public:

  EFAPI_Real(void) :
  Beta(0),
  X( N_Len, 0 ),
  W( N_Len, AVX_Vector<ProcType>(R_Len, 0) ),
  R( R_Len, AVX_Vector<ProcType>(R_Len, 0) ),
  y( R_Len, 0 ),
  h( R_Len, 0 ),
  g( R_Len, 0 ),
  Z( R_Len, AVX_Vector<ProcType>(R_Len,0) ),
  EpsSqr( 0 ),
  Tau( 0 ),
  Eta( 0 ),
  yp( R_Len, 0 ),
  hp( R_Len, 0 ),
  Eps( R_Len, 0 ),
  ep( N_Len, 0 ),
  Vect_N_1( N_Len, 0 ),
  Vect_R_1( R_Len, 0 ),
  Vect_R_2( R_Len, 0 ),
  Vect_R_3( N_Len, 0 ),
  Mat_N_R_1( N_Len, AVX_Vector<ProcType>(R_Len, 0) ),
  Var_1(0)
  {
  }

  EFAPI_Real( const ProcType& BetaParam ) :
  Beta(BetaParam),
  X( N_Len, 0 ),
  W( N_Len, AVX_Vector<ProcType>(R_Len, 0) ),
  R( R_Len, AVX_Vector<ProcType>(R_Len, 0) ),
  y( R_Len, 0 ),
  h( R_Len, 0 ),
  g( R_Len, 0 ),
  Z( R_Len, AVX_Vector<ProcType>(R_Len,0) ),
  EpsSqr( 0 ),
  Tau( 0 ),
  Eta( 0 ),
  yp( R_Len, 0 ),
  hp( R_Len, 0 ),
  Eps( R_Len, 0 ),
  ep( N_Len, 0 ),
  Vect_N_1( N_Len, 0 ),
  Vect_R_1( R_Len, 0 ),
  Vect_R_2( R_Len, 0 ),
  Vect_R_3( N_Len, 0 ),
  Mat_N_R_1( N_Len, AVX_Vector<ProcType>(R_Len, 0) ),
  Var_1(0)
  {
    this->Init(BetaParam);
  }

  void  Init( const ProcType& BetaParam )
  {
    this->Beta  = BetaParam;

    // Init W
    for( auto K=0U ; K < N_Len ; ++K )
    {
      for( auto L=0U ; L < R_Len ; ++L )
      {
        this->W[K][L]  = (K == L) ? 1 : 0;
      }
    }

    // Init Z
    for( auto K=0U ; K < R_Len ; ++K )
    {
      for( auto L=0U ; L < R_Len ; ++L )
      {
        this->Z[K][L]  = (K == L) ? 1 : 0;
      }
    }
  }

  uint32_t  GetDataLen(void) const
  {
    return  N_Len;
  }

  uint32_t  GetNumEigenvectors(void) const
  {
    return  R_Len;
  }

  /////////////////
  // Accessors
  /////////////////

  // Vector size is 'R_Len'
  const VectorType&  GetEigenVector( const uint32_t NumParam ) const
  {
    return  this->W[NumParam];
  }

  // Matrix size is 'N_Len * R_Len'
  const MatrixType&  GetEigenMatrix(void) const
  {
    return  this->W;
  }

  uint32_t  GetEigenMatrixSize(void) const
  {
    return  N_Len * R_Len;
  }

  const ProcType&  GetEigenValue( const uint32_t NumParam ) const
  {
    return  this->Z[NumParam][NumParam];
  }

  // Matrix size is 'R_Len * R_Len'
  const MatrixType&  GetEigenValueMatrix(void) const
  {
    return  this->Z;
  }

  uint32_t  GetEigenValueMatrixSize(void) const
  {
    return  R_Len * R_Len;
  }

  /////////////////
  // Processing
  /////////////////

  void  DataStep( const AVX_Vector<InType>& In )
  {
    assert( In.size() == N_Len );

    // Convert data to processing type
    for( auto K=0U ; K < N_Len ; ++K )
    {
      this->X[K] = In[K];
    }

    // [ 1]  y = W[h] * x
    this->MatVectMult_MhV_N( this->W, this->X, this->y );

    // [ 2]  h = Z * y
    this->MatVectMult_MV_N( this->Z, this->y, this->h );

    // [ 3]  g = h / ( beta + y[H] * h )
    assert( (this->Beta + DotProduct_VhV_N( this->y, this->h )) != 0. );

    this->VectScalarMult_VN_V
    (
      this->h,
      1. / ( this->Beta + DotProduct_VhV_N( this->y, this->h ) ),
      this->g
    );

    // [ 4]  EpsSqr = NormSqr(x) - NormSqr(y)
    this->EpsSqr = this->VectMag2_VhV_N(this->X) - this->VectMag2_VhV_N(this->y);

    // [ 5]  Tau = epsSqr / (1 + EpsSqr * Norm2(g) + sqrt( 1 + EpsSqr * Norm2(g)))
    const ProcType  Norm2G = this->VectMag2_VhV_N(this->g);
    this->Tau = this->EpsSqr / ( 1. + this->EpsSqr * Norm2G + sqrt( 1. + this->EpsSqr * Norm2G ));

    // [ 6]  Eta = 1 - Tau * Norm2(g)
    this->Eta = 1. - this->Tau * Norm2G;

    // [ 7]  Yp = Eta * y + Tau * g
    this->VectScalarMult_VN_V( this->y, this->Eta, this->Vect_R_1 );
    this->VectScalarMult_VN_V( this->g, this->Tau, this->Vect_R_2 );
    this->VectAdd_VV_V( this->Vect_R_1, this->Vect_R_2, this->yp );

    // [ 8]  hp = Z[H] * yp
    this->MatVectMult_MhV_N( this->Z, this->yp, this->hp );

    // [ 9]  Eps
    this->MatVectMult_MV_N( this->Z, this->g, this->Vect_R_1 );
    this->Var_1 = DotProduct_VhV_N( this->hp, this->g );
    this->VectScalarMult_VN_V( this->g, this->Var_1, this->Vect_R_2 );
    this->VectSub_VV_V( this->Vect_R_1, this->Vect_R_2, this->Eps );
    this->VectScalarMult_VN_X( this->Eps, (this->Tau/this->Eta) );

    // [10]  Z = (1/Beta) * ( Z - g * hp[H] + Eps * g[H] )
    this->VectOuterProdAndMatSub_VVh_Ms( this->g, this->hp, this->Z );
    this->VectOuterProdAndMatAdd_VVh_Ms( this->Eps, this->g, this->Z );
    this->MatScalarMult_MN_X( this->Z, (1./this->Beta) );

    // [11]  ep = Eta * x - W * yp
    this->VectScalarMult_VN_V( this->X, this->Eta, this->Vect_N_1 );
    this->MatVectMult_MV_N( this->W, this->yp, this->Vect_R_3 );
    this->VectSub_VV_V( this->Vect_N_1, this->Vect_R_3, this->ep );

    // [12]  W = W + ep * g[H]
    this->VectOuterProdAndMatAdd_VVh_Ms( this->ep, this->g, this->W );
  }

  //////////////////////
  // Signal Processing
  //////////////////////

  // Not Thread Safe with rest of its class
  // -Project into the spanned sub-space
  void  ProjectVectIntoSubSpace( const AVX_Vector<InType>&  SourceParam,
                                       AVX_Vector<InType>&  OutputParam  )
  {
    assert( SourceParam.size() == N_Len );
    assert( OutputParam.size() == N_Len );

    // Temp_1:  W[H] * SourceParam
    for( auto K=0U ; K < R_Len ; ++K )
    {
      this->Vect_R_1[K] = 0;

      for( auto L=0U ; L < N_Len ; ++L )
      {
        this->Vect_R_1[K] += this->W[L][K] * SourceParam[L];
      }
    }

    // Temp_2:  Temp_1 * W
    for( auto K=0U ; K < N_Len ; ++K )
    {
      OutputParam[K] = 0;

      for( auto L=0U ; L < R_Len ; ++L )
      {
        OutputParam[K] += this->W[K][L] * this->Vect_R_1[L];
      }
    }
  }

  // Project into the PERP spanned sub-space
  // -ProjectVectIntoSubSpace + ProjectVectIntoPerpSpace = Identity
  void  ProjectVectIntoPerpSpace( const AVX_Vector<InType>&  SourceParam,
                                        AVX_Vector<InType>&  OutputParam  )
  {
    assert( SourceParam.size() == N_Len );
    assert( OutputParam.size() == N_Len );

    this->ProjectVectIntoSubSpace( SourceParam, OutputParam  );

    // Calculate:  I - P
    for( auto K=0U ; K < N_Len ; ++K )
    {
      OutputParam[K] -= SourceParam[K];
    }
  }

private:

  // Parameters
  ProcType                          Beta;       // S

  // Variables
  AVX_Vector<ProcType>              X;          // N          (N == N_Len)
  AVX_Vector<AVX_Vector<ProcType>>  W;          // N * R      (R == R_Len)
  AVX_Vector<AVX_Vector<ProcType>>  R;          // R * R
  AVX_Vector<ProcType>              y;          // R
  AVX_Vector<ProcType>              h;          // R
  AVX_Vector<ProcType>              g;          // R
  AVX_Vector<AVX_Vector<ProcType>>  Z;          // R * R
  ProcType                          EpsSqr;     // S
  ProcType                          Tau;        // S
  ProcType                          Eta;        // S
  AVX_Vector<ProcType>              yp;         // R
  AVX_Vector<ProcType>              hp;         // R
  AVX_Vector<ProcType>              Eps;        // R
  AVX_Vector<ProcType>              ep;         // N

  // New working variables
  AVX_Vector<ProcType>              Vect_N_1;   // [*]  N
  AVX_Vector<ProcType>              Vect_R_1;   // [*]  R
  AVX_Vector<ProcType>              Vect_R_2;   // [*]  R
  AVX_Vector<ProcType>              Vect_R_3;   // [*]  N
  AVX_Vector<AVX_Vector<ProcType>>  Mat_N_R_1;  // [*]  N * R
  ProcType                          Var_1;      // [*]  S

private:

  ProcType  DotProduct_VhV_N( const AVX_Vector<ProcType>& U,
                              const AVX_Vector<ProcType>& V   ) const
  {
    assert( U.size() == V.size() );

    ProcType  Ret = 0;

    for( auto K=0U ; K < U.size() ; ++K )
    {
      Ret  += U[K] * V[K];
    }

    return  Ret;
  }

  ProcType  VectMag2_VhV_N( AVX_Vector<ProcType>& V ) const
  {
    ProcType  Ret = 0;

    for( auto K=0U ; K < V.size() ; ++K )
    {
      Ret  += pow( V[K], 2 );
    }

    return  Ret;
  }

  void  MatVectMult_MhV_N(
                         const AVX_Vector<AVX_Vector<ProcType>>&  Mat,
                         const AVX_Vector<ProcType>&          Vect,
                         AVX_Vector<ProcType>&          VectOut
                       ) const
  {
    assert( Mat.size() == Vect.size() );

    for( auto K=0U ; K < Mat[0].size() ; ++K )
    {
      VectOut[K] = 0;

      for( auto L=0U ; L < Mat.size() ; ++L )
      {
        VectOut[K] += Mat[L][K] * Vect[L];
      }
    }
  }

  void  MatVectMult_MV_N(
                         const AVX_Vector<AVX_Vector<ProcType>>&  Mat,
                         const AVX_Vector<ProcType>&          Vect,
                         AVX_Vector<ProcType>&          VectOut
                       ) const
  {
    assert( Mat[0].size() == Vect.size() );
    assert( Mat.size() == VectOut.size() );

    for( auto K=0U ; K < Mat.size() ; ++K )
    {
      VectOut[K] = 0;

      for( auto L=0U ; L < Mat[K].size() ; ++L )
      {
        VectOut[K] += Mat[K][L] * Vect[L];
      }
    }
  }

  void  MatMatAdd_MM_M(
                         const AVX_Vector<AVX_Vector<ProcType>>&  Mat1,
                         const AVX_Vector<AVX_Vector<ProcType>>&  Mat2,
                         AVX_Vector<AVX_Vector<ProcType>>&  MatOut
                       ) const
  {
    assert( Mat1.size() == Mat2.size() );
    assert( Mat1[0].size() == Mat2[0].size() );
    assert( Mat1.size() == MatOut.size() );
    assert( Mat1[0].size() == MatOut[0].size() );

    for( auto K=0U ; K < Mat1.size() ; ++K )
    for( auto L=0U ; L < Mat1[0].size() ; ++L )
    {
      MatOut[K][L] = Mat1[K][L] + Mat2[K][L];
    }
  }

  void  MatMatSub_MM_M(
                         const AVX_Vector<AVX_Vector<ProcType>>&  Mat1,
                         const AVX_Vector<AVX_Vector<ProcType>>&  Mat2,
                         AVX_Vector<AVX_Vector<ProcType>>&  MatOut
                       ) const
  {
    assert( Mat1.size() == Mat2.size() );
    assert( Mat1[0].size() == Mat2[0].size() );
    assert( Mat1.size() == MatOut.size() );
    assert( Mat1[0].size() == MatOut[0].size() );

    for( auto K=0U ; K < Mat1.size() ; ++K )
    for( auto L=0U ; L < Mat1[0].size() ; ++L )
    {
      MatOut[K][L] = Mat1[K][L] - Mat2[K][L];
    }
  }

  void  MatScalarMult_MN_X(
      AVX_Vector<AVX_Vector<ProcType>>&  Mat,
                         const ProcType&            Scalar
                       ) const
  {
    for( auto K=0U ; K < Mat.size() ; ++K )
    for( auto L=0U ; L < Mat[L].size() ; ++L )
    {
      Mat[K][L] *= Scalar;
    }
  }

  void  VectScalarMult_VN_V(
                         const AVX_Vector<ProcType>&  Vect,
                         const ProcType&          Scalar,
                         AVX_Vector<ProcType>&  VectOut
                       ) const
  {
    assert( Vect.size() == VectOut.size() );

    for( auto K=0U ; K < Vect.size() ; ++K )
    {
      VectOut[K] = Scalar * Vect[K];
    }
  }

  void  VectScalarMult_VN_X(
      AVX_Vector<ProcType>&  Vect,
                         const ProcType&          Scalar
                       ) const
  {
    for( auto K=0U ; K < Vect.size() ; ++K )
    {
      Vect[K] *= Scalar;
    }
  }

  void  VectSub_VV_V(
                      const AVX_Vector<ProcType>&  Vect1,
                      const AVX_Vector<ProcType>&  Vect2,
                      AVX_Vector<ProcType>&  VectOut
                    ) const
  {
    assert( Vect1.size() == Vect2.size() );
    assert( Vect1.size() == VectOut.size() );

    for( auto K=0U ; K < Vect1.size() ; ++K )
    {
      VectOut[K] = Vect1[K] - Vect2[K];
    }
  }

  void  VectAdd_VV_V(
                      const AVX_Vector<ProcType>&  Vect1,
                      const AVX_Vector<ProcType>&  Vect2,
                      AVX_Vector<ProcType>&  VectOut
                    ) const
  {
    assert( Vect1.size() == Vect2.size() );
    assert( Vect1.size() == VectOut.size() );

    for( auto K=0U ; K < Vect1.size() ; ++K )
    {
      VectOut[K]  = Vect1[K] + Vect2[K];
    }
  }

  void  VectOuterProd_VVh_M(
                      const AVX_Vector<ProcType>&          Vect1,
                      const AVX_Vector<ProcType>&          Vect2,
                      AVX_Vector<AVX_Vector<ProcType>>&  Out
                    ) const
  {
    assert( Vect1.size() == Out.size() );
    assert( Vect2.size() == Out[0].size() );

    for( auto K=0U ; K < Vect1.size() ; ++K )
    for( auto L=0U ; L < Vect2.size() ; ++L )
    {
      Out[K][L]  = Vect1[K] * Vect2[L];
    }
  }

  void  VectOuterProdAndMatSub_VVh_Ms(
                      const AVX_Vector<ProcType>&         Vect1,
                      const AVX_Vector<ProcType>&         Vect2,
                      AVX_Vector<AVX_Vector<ProcType>>&   Out
                    ) const
  {
    assert( Vect1.size() == Out.size() );
    assert( Vect2.size() == Out[0].size() );

    for( auto K=0U ; K < Vect1.size() ; ++K )
    for( auto L=0U ; L < Vect2.size() ; ++L )
    {
      Out[K][L]  -= Vect1[K] * Vect2[L];
    }
  }

  void  VectOuterProdAndMatAdd_VVh_Ms(
                      const AVX_Vector<ProcType>&         Vect1,
                      const AVX_Vector<ProcType>&         Vect2,
                      AVX_Vector<AVX_Vector<ProcType>>&   Out
                    ) const
  {
    assert( Vect1.size() == Out.size() );
    assert( Vect2.size() == Out[0].size() );

    for( auto K=0U ; K < Vect1.size() ; ++K )
    for( auto L=0U ; L < Vect2.size() ; ++L )
    {
      Out[K][L]  += Vect1[K] * Vect2[L];
    }
  }
};


