///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  FIFO.h                                                                            //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#pragma once

#include  <mutex>
#include  <condition_variable>
#include  <list>
#include  <atomic>
#include  <iostream>
#include  <iomanip>
#include  "AlignedMemAlloc.h"
#include  "BufMap.h"
#include  "Exception.h"

using namespace std;

////////////////////////////////////////////////////////////////////////////////////////////////////
//  FIFO Addressing
////////////////////////////////////////////////////////////////////////////////////////////////////
//
// - FIFOs are defined by two N-Dimensional vectors:
//   - Extents (size of FIFO), e.g. [ 1024  2  128 ].
//   - Wrap flags: if TRUE the dimension wraps about the end point, e.g. [ true  false  false ].
// - FIFO locations are uniquely defined by a N-dimensional vector, e.g. [ X Y Z ].
// - Entries in a FIFO are memory contiguous over their rightmost entry (then moving leftward):
//   - e.g. [ 0 0 0 ], [ 0 0 1 ], ..., [ 0 1 0 ], [ 0 1 1 ], ..., [ 1 0 0 ], [ 1 0 1 ].
// - The left-most entry is special and is ALWAYS the FIFO Sample Number (FIFO column number):
//   - e.g. [ FifoSampleNumber ... ]
// - Other dimensions are at the discretion of the programmer and defined by the Extents vector.
//
////////////////////////////////////////////////////////////////////////////////////////////////////


// For debug prints
extern  mutex  DebugMutex;     // Global Debug Mutex


template<typename DataType,typename AllocType>
class  FIFO_Reader;

template<typename DataType,typename AllocType>
class  FifoWriterTool;

///////////////////////////////////////////////////////////////////////////
//  FIFO<typename DataType,typename AllocType>
///////////////////////////////////////////////////////////////////////////

template<typename DataType,typename AllocType>
class  FIFO
{
public:

  static const uint32_t  SizeRatio = sizeof(DataType)/sizeof(AllocType);

  class  WriteLockTool
  {
  public:

    typedef struct sWriteSchema
    {
      uint32_t    Length;
      uint32_t    Start;
      uint32_t    End;

      // Note that StartWrapped is assumed 0
      uint32_t    EndWrapped;
      bool        WrapFlag;
    }   WriteSchemaType;

  private:

    // Inputs
    FIFO<DataType,AllocType>&   FifoRef;
    uint32_t                    NumData;    // Amount of data to lock for write

  public:

    WriteLockTool( FIFO<DataType,AllocType>& _FIFO ) :
    FifoRef(_FIFO),
    NumData(0)
    {
    }

    inline void  SetNumData( const uint32_t _NumData )
    {
      this->NumData = _NumData;
    }

    inline uint32_t  GetNumData(void) const
    {
      return  this->NumData;
    }

    inline bool  IsFifoWrap(void) const
    {
      return  this->NumData > this->FifoRef.EndProcDiff();
    }

    inline FIFO<DataType,AllocType>&  GetFifoRef(void)
    {
      return  this->FifoRef;
    }

    inline const FIFO<DataType,AllocType>&  GetFifoRef(void) const
    {
      return  this->FifoRef;
    }

    /////////////////////////////
    // Physical write locations
    /////////////////////////////

    DataType*  Loc_Base0( const vector<uint32_t>& _LocationVector, const WriteSchemaType& _WST ) const
    {
      return  &this->FifoRef.Data[Offset_WriteStart0(_LocationVector,_WST)];
    }

    DataType*  Loc_Base1( const vector<uint32_t>& _LocationVector ) const
    {
      return  &this->FifoRef.Data[Offset_WriteStart1(_LocationVector)];
    }

    DataType*  Loc_End0( const vector<uint32_t>& _LocationVector, const WriteSchemaType& _WST ) const
    {
      return  &this->FifoRef.Data[Offset_WriteEnd0(_LocationVector,_WST)];
    }

    DataType*  Loc_End1( const vector<uint32_t>& _LocationVector, const WriteSchemaType& _WST ) const
    {
      return  &this->FifoRef.Data[Offset_WriteEnd1(_LocationVector,_WST)];
    }

    uint32_t  Loc_Num0( const vector<uint32_t>& _LocationVector, const WriteSchemaType& _WST ) const
    {
      return  Loc_End0( _LocationVector, _WST ) - Loc_Base0( _LocationVector, _WST );
    }

    uint32_t  Loc_Num1( const vector<uint32_t>& _LocationVector, const WriteSchemaType& _WST ) const
    {
      return  Loc_End1( _LocationVector, _WST ) - Loc_Base1( _LocationVector );
    }

    uint32_t  Offset_WriteStart0( vector<uint32_t> _LocationVector, const WriteSchemaType& _WST ) const
    {
      _LocationVector[0] = _WST.Start;
      return  this->FifoRef.Extent.Offset(_LocationVector);
    }

    uint32_t  Offset_WriteEnd0( vector<uint32_t> _LocationVector, const WriteSchemaType& _WST ) const
    {
      _LocationVector[0] = _WST.End;
      return  this->FifoRef.Extent.Offset(_LocationVector);
    }

    uint32_t  Offset_WriteStart1( vector<uint32_t> _LocationVector ) const
    {
      _LocationVector[0] = 0;
      return  this->FifoRef.Extent.Offset(_LocationVector);
    }

    uint32_t  Offset_WriteEnd1( vector<uint32_t> _LocationVector, const WriteSchemaType& _WST ) const
    {
      _LocationVector[0] = _WST.EndWrapped;
      return  this->FifoRef.Extent.Offset(_LocationVector);
    }

    DataType*  GetBaseAddressOfFIFO(void) const
    {
      return  &this->FifoRef.Data[0];
    }

    FIFO<DataType,AllocType>&  GetFifo_Ref(void)
    {
      return  this->FifoRef;
    }

    void  GetWriteSchema( WriteSchemaType& _WriteSchema )
    {
      _WriteSchema.Length   = this->NumData;
      _WriteSchema.Start    = this->FifoRef.Proc;

      const int32_t NewEnd = ( _WriteSchema.Start + this->NumData );

      _WriteSchema.WrapFlag = NewEnd > (int) this->FifoRef.Extent.GetLengthLowestDim();

      if( !_WriteSchema.WrapFlag )
      { // No wrap.  A single buffer
        _WriteSchema.End        = NewEnd;
        _WriteSchema.EndWrapped = 0;
      }
      else
      { // Wrap around end of FIFO
        _WriteSchema.End        = this->FifoRef.Extent.GetLengthLowestDim();
        _WriteSchema.EndWrapped = ( NewEnd - this->FifoRef.Extent.GetLengthLowestDim() );
      }
    }

    string  Display(void) const
    {
      ostringstream   oS;
      oS << this->FifoRef.Display();
      oS << "  WWrap:" << boolalpha << this->IsFifoWrap();
      oS << endl;
      return    oS.str();
    }

    string  DisplayWriteSchema( const WriteSchemaType& _WriteSchema ) const
    {
      ostringstream   oS;

      if(_WriteSchema.WrapFlag)
      {
        oS << "-WriteSchema:"
            << "  Start:" << setw(3)  << _WriteSchema.Start
            << "  End+1:" << setw(3)  << _WriteSchema.End
            << "  Start2:" << setw(3) << 0
            << "  End2+1:" << setw(3) << _WriteSchema.EndWrapped
            << "  Len:" << setw(3)    << _WriteSchema.Length
            << "  Wrapped:" << boolalpha << _WriteSchema.WrapFlag;
      }
      else
      {
        oS << "-WriteSchema:"
            << "  Start:" << setw(3) << _WriteSchema.Start
            << "  End+1:" << setw(3) << _WriteSchema.End
            << "  Len:" << setw(3)   << _WriteSchema.Length
            << "  Wrapped:" << boolalpha << _WriteSchema.WrapFlag;
      }

      return    oS.str();
    }
  };

public:

  FIFO( const BufMap<DataType,AllocType>& _Extent ) :
  Extent {_Extent},
  Data( this->Extent.GetNumElements() * FIFO::SizeRatio ),
  Front {0},
  Proc {0},
  Back {0},
  NumReadable {0},
  NumWritable {0},
  NumFree {this->Extent.GetLengthLowestDim()},
  KillFlag {false},
  EndOfDataFlag(false)
  {
    lock_guard<mutex>  Lock(this->MainMut);

    assert( sizeof(DataType) >= sizeof(AllocType) );

    // Since this is a FIFO, we assume we can wrap on LSD
    this->Extent.SetLowestWrapFlag(true);
  }

  // Call with Mutex 'MainMut' taken
  void  RegisterListener_L( FIFO_Reader<DataType,AllocType>* _ActiveReader  )
  {
    lock_guard<mutex>  Lock(this->MainMut);

    this->ActiveReaders.remove(_ActiveReader);
    this->ActiveReaders.push_back(_ActiveReader);
  }

  // Call with Mutex 'MainMut' taken
  void  RegisterListener( FIFO_Reader<DataType,AllocType>* _ActiveReader  )
  {
    this->ActiveReaders.remove(_ActiveReader);
    this->ActiveReaders.push_back(_ActiveReader);
  }

  // Call with Mutex 'MainMut' taken
  void  UnRegisterListener( FIFO_Reader<DataType,AllocType>* _ActiveReader )
  {
    lock_guard<mutex>   Lock(this->MainMut);

    this->ActiveReaders.remove(_ActiveReader);
  }

  void  SetKillFlag( const bool _KillFlag=true )
  {
    this->KillFlag = _KillFlag;

    if(_KillFlag)
    {
      for( auto I=this->ActiveReaders.begin() ; I != this->ActiveReaders.end() ; ++I )
      {
        (*I)->SetKillFlag(true);
      }

      this->CV.notify_all();
    }
  }

  uint32_t  GetVectorLen(void) const
  {
    return  Extent.GetDimensions().size();
  }

  //////////////////////////////////////////////////////////////////////////////////////
  // Data Stream Flags
  //////////////////////////////////////////////////////////////////////////////////////

  void  SetEndOfDataFlag(void)
  {
    this->EndOfDataFlag = true;
  }

  bool  GetEndOfDataFlag(void) const
  {
    return  this->EndOfDataFlag;
  }

  bool  IsFifoComplete(void) const
  {
    return  this->EndOfDataFlag && (this->GetNumReadable() == 0U);
  }

  //////////////////////////////////////////////////////////////////////////////////////
  // Accessors
  //////////////////////////////////////////////////////////////////////////////////////

  condition_variable&  GetCV(void)
  {
    return  this->CV;
  }

  mutex&  GetMut(void)
  {
    return  this->MainMut;
  }

  const BufMap<DataType,AllocType>&  GetBufMap(void) const
  {
    return  this->Extent;
  }

  uint32_t  GetNumReadable(void) const
  {
    return  this->NumReadable;
  }

  uint32_t  GetNumWritable(void) const
  {
    return  this->NumWritable;
  }

  uint32_t  GetNumFree(void) const
  {
    return  this->NumFree;
  }

  uint32_t  GetBack(void) const
  {
    return  this->Back;
  }

  uint32_t  GetProc(void) const
  {
    return  this->Proc;
  }

  const list<FIFO_Reader<DataType,AllocType>*>&  GetActiveReaders(void) const
  {
    return  this->ActiveReaders;
  }

  const AVX_Vector<AllocType>&  GetData(void) const
  {
    return  this->Data;
  }

private:

  //////////////////////////////////////////////////////////////////////////////////////
  // Debug modes
  //////////////////////////////////////////////////////////////////////////////////////

  const bool  DebugPrint = false;

  //////////////////////////////////////////////////////////////////////////////////////
  // Data description and allocation
  //////////////////////////////////////////////////////////////////////////////////////

  BufMap<DataType,AllocType>  Extent;
  AVX_Vector<AllocType>       Data;

  //////////////////////////////////////////////////////////////////////////////////////
  // These Pointers are the Least Significant Digit { A, ..., Front|Back|Proc };
  //    Next Location to Write Into   ---   Front <= Proc <= Back (wrap)
  //////////////////////////////////////////////////////////////////////////////////////

  // The location of the last item written - 1
  uint32_t      Front;

  // The location of the last item available for read - 1
  uint32_t      Proc;

  // The location of the first item available for read
  uint32_t      Back;

  // Number of items that are readable in the least significant dimension
  uint32_t      NumReadable;

  // Number of items that are writable in the least significant dimension
  uint32_t      NumWritable;

  // Number of items that are not being used in the least significant dimension
  uint32_t      NumFree;

  //////////////////////////////////////////////////////////////////////////////////////
  // Synchronization and control
  //////////////////////////////////////////////////////////////////////////////////////

  atomic<bool>          KillFlag;
  mutex                 MainMut;
  condition_variable    CV;
  bool                  EndOfDataFlag;

  //////////////////////////////////////////////////////////////////////////////////////
  // Active readers
  //////////////////////////////////////////////////////////////////////////////////////

  list<FIFO_Reader<DataType,AllocType>*>  ActiveReaders;

public:

  //////////////////////////////////////////////////////////////////////////////////////
  // Data copies
  //////////////////////////////////////////////////////////////////////////////////////

  // General FIFO writable pointer at location '_Loc'
  DataType*  GetFifoAddress
        (
          const FIFO<DataType,AllocType>::WriteLockTool&                            _WriteLockTool,
          const typename FIFO<DataType,AllocType>::WriteLockTool::WriteSchemaType&  _WriteSchema,
          const vector<uint32_t>&                                                   _Loc
        )
  {
    return  _WriteLockTool.Loc_Base0( _Loc, _WriteSchema );
  }

  //////////////////////////////////////////////////////////////////////////////////////
  // Write Requests
  //////////////////////////////////////////////////////////////////////////////////////

  bool  FIFO_Write_Start
        (
          const uint32_t                        AmountToWrite,
          FifoWriterTool<DataType,AllocType>&   FWT
        )
  {
    unique_lock<mutex>  Lock(this->MainMut);

    // Get the WriteLockTool
    auto&  WLT = FWT.GetWriteLockTool();

    // Save the desired transfer length
    WLT.SetNumData(AmountToWrite);

    // Wait on condition variable
    this->CV.wait( Lock, [&]{ return this->KillFlag || (this->NumFree >= AmountToWrite); } );

    Lock.unlock();

    if( this->KillFlag )  return false;

    this->Incr_Front_L_T( AmountToWrite );

    // Just tickle the CV of this FIFO
    this->NotifyOwnCV();

    // Generate and cache the write schema
    FWT.GetWriteSchema_Calc();

    return this->KillFlag;
  }

  void  FIFO_Write_End( FifoWriterTool<DataType,AllocType>& FWT )
  {
    // Get the WriteLockTool
    const auto&  WLT = FWT.GetWriteLockTool();

    this->Incr_Proc_L_T( WLT.GetNumData() );

    // Tickle the CV of this FIFO and the CV of all the reader FIFOs
    this->NotifyAllReadersCVs();
  }

  //////////////////////////////////////////////////////////////////////////////////////
  // Utility
  //////////////////////////////////////////////////////////////////////////////////////

  inline uint32_t  Normalize( const uint32_t _Value1 ) const
  {
    return  _Value1 % this->GetBufMap().GetLengthLowestDim();
  }

  // Control Variable (CV) ticklers
  inline void  NotifyAllReadersCVs(void)
  {
    for( auto I: this->ActiveReaders )
    {
      I->GetCV().notify_all();
    }
  }

  inline void  NotifyOwnCV(void)
  {
    this->CV.notify_all();
  }

  //////////////////////////////////////////////////////////////////////////////////////
  // Incr
  //////////////////////////////////////////////////////////////////////////////////////
  // Suffixes:
  //      L:  Lock
  //      T:  Test State
  /////////////////////////////////////////

  void  Incr_Front_L_T( const uint32_t _IncrBy )
  {
    this->Incr_Front_L(_IncrBy);

    // Look for exception conditions
    if(this->DebugPrint)
    {
      ostringstream   oS;
      oS << "IncrFront(" << _IncrBy << ")";
      this->TestForException(oS.str());
    }
  }

  inline void  Incr_Front( const uint32_t _IncrBy )
  {
    assert( this->NumFree >= _IncrBy );

    this->NumFree      -= _IncrBy;
    this->NumWritable  += _IncrBy;
    this->Front         = this->Normalize( this->Front + _IncrBy );
  }

  inline void  Incr_Front_L( const uint32_t _IncrBy )
  {
    lock_guard<mutex>   Lock(this->MainMut);

    this->Incr_Front(_IncrBy);
  }

  void  Incr_Proc_L_T( const uint32_t _IncrBy )
  {
    {
      lock_guard<mutex>  Lock(this->MainMut);

      // Lock all readers
      for( auto I: this->ActiveReaders )
      {
        I->GetMut().lock();
      }

      // Update all:  RFIFO.Proc = FIFO.Proc
      if( _IncrBy > 0 )
      {
        this->Incr_Proc(_IncrBy);

        for( auto I: this->ActiveReaders )
        {
          I->Incr_Proc(_IncrBy);
        }
      }

      // UnLock all readers
      for( auto I: this->ActiveReaders )
      {
        I->GetMut().unlock();
      }
    }

    // Look for exception conditions
    if(this->DebugPrint)
    {
      ostringstream   oS;
      oS << "IncrProc(" << _IncrBy << ")";
      this->TestForException(oS.str());
    }
  }

  inline void  Incr_Proc( const uint32_t _IncrBy )
  {
    assert( this->NumWritable >= _IncrBy );

    this->NumReadable  += _IncrBy;
    this->NumWritable  -= _IncrBy;
    this->Proc          = this->Normalize( this->Proc + _IncrBy );
  }

  inline void  Incr_Proc_L( const uint32_t _IncrBy )
  {
    lock_guard<mutex>   Lock(this->MainMut);

    this->Incr_Proc(_IncrBy);
  }

  void  Incr_Back_L_T( const uint32_t _IncrBy )
  {
    this->Incr_Back_L(_IncrBy);

    // Look for exception conditions
    if(this->DebugPrint)
    {
      ostringstream   oS;
      oS << "IncrBack(" << _IncrBy << ")";
      this->TestForException(oS.str());
    }
  }

  inline void  Incr_Back_L( const uint32_t _IncrBy )
  {
    lock_guard<mutex>   Lock(this->MainMut);

    this->Incr_Back(_IncrBy);
  }

  inline void  Incr_Back( const uint32_t _IncrBy )
  {
    assert( this->NumReadable >= _IncrBy );

    this->NumFree      += _IncrBy;
    this->NumReadable  -= _IncrBy;
    this->Back          = this->Normalize( this->Back + _IncrBy );
  }

  //////////////////////////////////////////////////////////////////////////////////////
  // Display
  //////////////////////////////////////////////////////////////////////////////////////

  string  Display( const bool _DisplayListeners=false ) const
  {
    ostringstream  oS;

    oS << "FIFO  [" << this << "]:"
       << "  F:" << this->Front
       << "  P:" << this->Proc
       << "  B:" << this->Back
       << "  nReadable:" << this->NumReadable
       << "  nWritable:" << this->NumWritable
       << "  nFree:" << this->NumFree
       << "  nLen:" << this->Extent.GetLengthLowestDim()
       << "  EODF:" << boolalpha << this->EndOfDataFlag;

    if(this->KillFlag)
    {
      oS << "  <KILL>";
    }

    oS << endl;

    if( _DisplayListeners )
    {
      for( auto I: this->ActiveReaders )
      {
        oS << "+" << I->Display();
      }
    }

    return    oS.str();
  }

public:

  inline uint32_t   EndProcDiff(void) const
  {
    return  this->Extent.LengthLowestDim - this->Proc;
  }

  inline uint32_t   EndBackDiff(void) const
  {
    return  this->Extent.LengthLowestDim - this->Back;
  }

  inline uint32_t   EndFrontDiff(void) const
  {
    return  this->Extent.LengthLowestDim - this->Front;
  }

  // Notes on write transfers into FIFOs.  It is necessary to restrict the valid
  //   write operations due to certain pathological cases.  In essence, we want to
  //   perform a write into an available section of the output FIFO.  We exclude
  //   operations that wrap around the FIFO more than once as this isn't a data
  //   transfer technique that is seen as useful.  Transfers that wrap around the
  //   end once are valid.

  template<class InType>
  void  WriteVectorToFifo
        (
          const InType*             InVect,
          const vector<uint32_t>&   FifoStart,
          const vector<uint32_t>&   FifoIncr,
          const uint32_t            TransferLength
        )
  {
    // Offsets into the FIFO
    const auto  First = this->Extent.Offset(FifoStart);
    const auto  Incr  = this->Extent.Offset(FifoIncr);

    // Look for the special case where the increment is one data item.
    //   In this case we do one contiguous transfer rather than multiple
    //   single-datum transfers.
    if( Incr == 1 )
    { // Data items are adjacent in memory --- faster transfers
      const uint32_t  LastPlusOne = First + TransferLength;

      if( LastPlusOne <= this->Extent.GetNumElements() )
      { // One 'std::copy' operation
        std::copy
        (
          &InVect[0],
          &InVect[TransferLength],
          &this->Data[First]
        );
      }
      else
      { // Two 'std::copy' operations wrapping around the end of the FIFO
        const uint32_t  DistToTopOfFifo = this->Extent.GetNumElements() - First;

        std::copy
        (
          &InVect[0],
          &InVect[DistToTopOfFifo],
          &this->Data[First]
        );

        // Note that this operation is always valid due to the IF test above
        std::copy
        (
          &InVect[DistToTopOfFifo],
          &InVect[TransferLength],
          &this->Data[First+DistToTopOfFifo]
        );
      }
    }
    else
    { // Data items are not adjacent in memory --- slower transfers
      uint32_t  Curr = First;

      for( auto K=0U ; K < TransferLength ; ++K )
      {
        this->Data[Curr] = InVect[K];

        // Incr 'Curr'
        if( (K+1) < TransferLength )
        {
          Curr += Incr;

          if( Curr >= this->Extent.GetNumElements() )
          {
            Curr -= this->Extent.GetNumElements();
            assert(Curr < this->Extent.GetNumElements());
          }
        }
      }
    }
  }

  inline void  WriteValueToFifo( const DataType& NewData, const uint32_t FifoPos )
  {
    this->Data[FifoPos] = NewData;
  }

  inline uint32_t  IncrFifoPos( const uint32_t FifoPos, const uint32_t FifoIncr ) const
  {
    uint32_t  Ret = FifoPos + FifoIncr;

    if( Ret > this->Extent.GetNumElements() )
    {
      Ret -= this->Extent.GetNumElements();
      assert( Ret < this->Extent.GetNumElements() );
    }

    return  Ret;
  }

private:

  void    TestForException( const string& _Tag )
  {
    lock_guard<mutex>  Lock(this->MainMut);

    if( (this->NumReadable+this->NumWritable+this->NumFree) != this->Extent.GetLengthLowestDim() )
    {
      cout << endl
           << "*Error '" << _Tag
           << "': '(this->NumReadable+this->NumWritable+this->NumFree) != this->Extent.GetLengthLowestDim()'" << endl
           << this->Display(true);
      cout.flush();
    }

    if( (Normalize(this->Back + this->NumReadable) != this->Proc ) )
    {
      cout << endl
           << "*Error '" << _Tag
           << "': '(Normalize(this->Back + this->NumWritable) != this->Proc )'" << endl
           << this->Display(true);
      cout.flush();
    }

    if( (Normalize(this->Proc + this->NumWritable) != this->Front ) )
    {
      cout << endl
           << "*Error '" << _Tag
           << "': '(Normalize(this->Proc + this->NumReadable) != this->Front )'" << endl
           << this->Display(true);
      cout.flush();
    }

    if( (Normalize(this->Front + this->NumFree) != this->Back ) )
    {
      cout << endl
           << "*Error '" << _Tag
           << "': '(Normalize(this->Front + this->NumFree) != this->Back )'" << endl
           << this->Display(true);
      cout.flush();
    }
  }
};


