///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  XML_Storage.h                                                                     //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#pragma once

#include      <string>
#include      <sstream>
#include      <random>
#include      "XML_Node.h"
#include      "Outputs.h"

using namespace std;


namespace   wej_parser
{
///////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////
// XML_Storage
///////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////

class     XML_Storage
{
private:

  // Parameters
  static const bool                 UseCompactXML_Format = false;

  // Binary file control bytes
  static const unsigned char        SOF_Byte_0, SOF_Byte_1, SOF_Byte_2, SOF_Byte_3;
  static const unsigned char        EOF_Byte_0, EOF_Byte_1, EOF_Byte_2, EOF_Byte_3;
  static const unsigned char        RandomSeed;

  typedef     linear_congruential_engine<unsigned char,179,131,0>   F_RAND;

  bool                              Valid;
  XML_Node                          Head;
  list<XML_Node::AttributeRec>      XmlDecs;

protected:

  bool    ParseNode
          (
            XML_Node&                         NodeParam,
            list<OutputRec>::const_iterator&  I,
            const Outputs&                    OuputsParam,
            const bool                        AtHead
          );

  bool    WriteBinLevel
          (
            const XML_Node&   NodeParam,
            ostream&          Out,
            F_RAND&           Rand
          ) const;

  bool    ReadBinLevel
          (
            XML_Node&         NodeParam,
            istream&          In,
            F_RAND&           Rand
          );

  static void           ApplyRand( unsigned char* Buf, const unsigned int Length, F_RAND& Rand );
  static unsigned char  ApplyRand( unsigned char  Buf, F_RAND& Rand );
  static string         ApplyRand( const string& Buf, F_RAND& Rand );

public:

  XML_Storage(void);

  inline bool     IsValid(void) const
  {
    return    Valid;
  }

  inline void     MarkInvalid(void)
  {
    Valid   = false;
  }

  inline void     PushXmlDecsLabelValuePair( const string LabelParam, const string ValueParam )
  {
    this->XmlDecs.push_back( XML_Node::AttributeRec( LabelParam, ValueParam ) );
  }

  bool      ReadFromXML_File( const string& XML_FilePath );

  bool      ReadFromBinFile( const string& ScenFilePath );

  bool      WriteToXML_File( const string& FilePath ) const;

  bool      WriteToBinFile( const string& FilePath ) const;

  string    GetXML_AsSourceString( const bool Compact ) const;

  string    Display(void) const;

  /////////////////////////////
  // LookUp Node
  /////////////////////////////

  XML_Node*         LookUpNodeWithException( const string& XmlPath )
  {
    return    this->Head.LookUpNodeWithException( XmlPath );
  }

  const XML_Node*   LookUpNodeWithException( const string& XmlPath ) const
  {
    return    this->Head.LookUpNodeWithException( XmlPath );
  }

  XML_Node*         LookUpNode( const string& XmlPath )
  {
    return    this->Head.LookUpNode( XmlPath );
  }

  const XML_Node*   LookUpNode( const string& XmlPath ) const
  {
    return    this->Head.LookUpNode( XmlPath );
  }

  /////////////////////////////
  // XML Decl Browsing
  /////////////////////////////

  typedef list<XML_Node::AttributeRec>::iterator          XmlDeclIterator;
  typedef list<XML_Node::AttributeRec>::const_iterator    XmlDeclConstIterator;

  inline list<XML_Node::AttributeRec>::iterator           XmlDeclBegin(void)
  {
    return    XmlDecs.begin();
  }

  inline list<XML_Node::AttributeRec>::iterator           XmlDeclEnd(void)
  {
    return    XmlDecs.end();
  }

  inline list<XML_Node::AttributeRec>::const_iterator     XmlDeclBegin(void) const
  {
    return    XmlDecs.begin();
  }

  inline list<XML_Node::AttributeRec>::const_iterator     XmlDeclEnd(void) const
  {
    return    XmlDecs.end();
  }

  /////////////////////////////
  // Get Node Head
  /////////////////////////////

  inline XML_Node&          GetHead(void)
  {
    return    Head;
  }

  inline const XML_Node&    GetHead(void) const
  {
    return    Head;
  }
};
}   // End namespace 'wej_parser'


